#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis


"""Classes and methods that control the performance assessment."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from pelicun import base, file_io, model, uq
from pelicun.__init__ import __version__ as pelicun_version  # type: ignore
from pelicun.base import EDP_to_demand_type, get

if TYPE_CHECKING:
    from pelicun.base import Logger

default_damage_processes = {
    'FEMA P-58': {
        '1_excessive.coll.DEM': {'DS1': 'collapse_DS1'},
        '2_collapse': {'DS1': 'ALL_NA'},
        '3_excessiveRID': {'DS1': 'irreparable_DS1'},
        '4_irreparable': {'DS1': 'ALL_NA'},
        '5_irreparable': {'DS1': 'collapse_DS0'},
    },
    # TODO(AZ): expand with ground failure logic
    'Hazus Earthquake - Lifeline Facilities': {
        '1_LF': {'DS5': 'collapse_DS1'},
        '2_collapse': {'DS1': 'ALL_NA'},
    },
    'Hazus Earthquake - Buildings': {
        '1_STR': {'DS5': 'collapse_DS1'},
        '2_excessive.coll.DEM': {'DS1': 'collapse_DS1'},
        '3_collapse': {'DS1': 'ALL_NA'},
        '4_excessiveRID': {'DS1': 'irreparable_DS1'},
        '5_irreparable': {'DS1': 'ALL_NA'},
        '6_irreparable': {'DS1': 'collapse_DS0'},
    },
    'Hazus Hurricane': {},
}


class AssessmentBase:
    """
    Base class for Assessment objects.

    Assessment objects manage the models, data, and calculations in pelicun.

    """

    __slots__: list[str] = [
        'asset',
        'damage',
        'demand',
        'log',
        'loss',
        'options',
        'stories',
        'unit_conversion_factors',
    ]

    def __init__(self, config_options: dict[str, Any] | None = None) -> None:
        """
        Initialize an Assessment object.

        Parameters
        ----------
        config_options:
            User-specified configuration dictionary.

        """
        self.stories: int | None = None
        self.options = base.Options(config_options, self)
        self.unit_conversion_factors: dict = base.parse_units(
            self.options.units_file
        )

        self.log: Logger = self.options.log
        self.log.msg(
            f'pelicun {pelicun_version} | \n',
            prepend_timestamp=False,
            prepend_blank_space=False,
        )
        self.log.print_system_info()
        self.log.div()
        self.log.msg('Assessment Started')

        self.demand: model.DemandModel = model.DemandModel(self)
        self.asset: model.AssetModel = model.AssetModel(self)
        self.damage: model.DamageModel = model.DamageModel(self)
        self.loss: model.LossModel = model.LossModel(self)

    @property
    def bldg_repair(self) -> model.LossModel:
        """
        Exists for <backwards compatibility>.

        Returns
        -------
        model.LossModel
            The loss model.

        """
        self.log.warning(
            '`.bldg_repair` is deprecated and will be dropped in '
            'future versions of pelicun. '
            'Please use `.loss` instead.'
        )

        return self.loss

    @property
    def repair(self) -> model.LossModel:
        """
        Exists for <backwards compatibility>.

        Returns
        -------
        RepairModel_DS
            The damage state-driven component loss model.

        """
        self.log.warning(
            '`.repair` is deprecated and will be dropped in '
            'future versions of pelicun. '
            'Please use `.loss` instead.'
        )
        return self.loss

    def get_default_data(
        self, method_name: str, model_type: str | None = None
    ) -> pd.DataFrame:
        """
        Load a default data file.

        Loads a default data file by name and returns it. This method
        is specifically designed to access predefined CSV files from a
        structured directory path related to the SimCenter fragility
        library.

        Parameters
        ----------
        method_name: str
            The name of the method to be used. This name is used to look
            up the full path to the model files in the Damage and Loss Model
            Library.
        model_type: str
            The type of model requested. Currently, the following types
            are supported: 'fragility', 'consequence_repair',
            'loss_repair'

        Returns
        -------
        pd.DataFrame
            The DataFrame containing the data loaded from the
            model CSV file.

        """
        # <backwards compatibility>
        if model_type is None:
            # Legacy inputs will have a filename provided instead of a
            # method name
            data_path = file_io.substitute_default_path(
                [f'PelicunDefault/{method_name}.csv'], log=self.log
            )[0]

        else:
            data_path = file_io.substitute_default_path(
                [f'PelicunDefault/{method_name}/{model_type}.csv'], log=self.log
            )[0]
        assert isinstance(data_path, str)

        data = file_io.load_data(
            data_path, None, orientation=1, reindex=False, log=self.log
        )

        assert isinstance(data, pd.DataFrame)
        return data

    def get_default_metadata(
        self, method_name: str, model_type: str | None = None
    ) -> dict:
        """
        Load a default metadata file and pass it to the user.

        Parameters
        ----------
        method_name: string
            The name of the method to be used. This name is used to look
            up the full path to the model files in the Damage and Loss Model
            Library.
        model_type: str
            The type of model requested. Currently, the following types
            are supported: 'fragility', 'consequence_repair',
            'loss_repair'

        Returns
        -------
        dict
            Default metadata describing the models available for the
            specified method.

        """
        # <backwards compatibility>
        if model_type is None:
            # Legacy inputs will have a filename provided instead of a
            # method name
            data_path = file_io.substitute_default_path(
                [f'PelicunDefault/{method_name}.json'], log=self.log
            )[0]

        else:
            data_path = file_io.substitute_default_path(
                [f'PelicunDefault/{method_name}/{model_type}.json'], log=self.log
            )[0]
        assert isinstance(data_path, str)

        with Path(data_path).open(encoding='utf-8') as f:
            data = json.load(f)

        return data  # noqa: RET504

    def calc_unit_scale_factor(self, unit: str) -> float:
        """
        Determine unit scale factor.

        Determines the scale factor from input unit to the
        corresponding base unit.

        Parameters
        ----------
        unit: str
            Either a unit name, or a quantity and a unit name
            separated by a space.
            For example: 'ft' or '100 ft'.

        Returns
        -------
        float
            Scale factor that convert values from unit to base unit

        Raises
        ------
        KeyError
            When an invalid unit is specified

        """
        unit_lst = unit.strip().split(' ')

        # check if there is a quantity specified; if yes, parse it
        if len(unit_lst) > 1:
            unit_count_str, unit_name = unit_lst
            unit_count = float(unit_count_str)

        else:
            unit_count = 1
            unit_name = unit_lst[0]

        try:
            scale_factor = unit_count * self.unit_conversion_factors[unit_name]

        except KeyError as exc:
            msg = f'Specified unit not recognized: {unit_count} {unit_name}'
            raise KeyError(msg) from exc

        return scale_factor

    def scale_factor(self, unit: str | None) -> float:
        """
        Get scale factor of given unit.

        Returns the scale factor of a given unit. If the unit is
        unknown it raises an error. If the unit is None it returns
        1.00.

        Parameters
        ----------
        unit: str
            A unit name.

        Returns
        -------
        float
            Scale factor

        Raises
        ------
        ValueError
            If the unit is unknown.

        """
        if unit is not None:
            if unit in self.unit_conversion_factors:
                scale_factor = self.unit_conversion_factors[unit]

            else:
                msg = f'Unknown unit: {unit}'
                raise ValueError(msg)
        else:
            scale_factor = 1.0

        return scale_factor


class Assessment(AssessmentBase):
    """
    Assessment class.

    Has methods implementing a Scenario-Based assessment.

    """

    __slots__: list[str] = []

    def calculate_damage(
        self,
        num_stories: int,
        demand_config: dict,
        demand_data_source: str | dict,
        cmp_data_source: str | dict[str, pd.DataFrame],
        damage_data_paths: list[str | pd.DataFrame],
        dmg_process: dict | None = None,
        scaling_specification: dict | None = None,
        residual_drift_configuration: dict | None = None,
        collapse_fragility_configuration: dict | None = None,
        block_batch_size: int = 1000,
    ) -> None:
        """
        Calculate damage.

        Parameters
        ----------
        num_stories: int
            Number of stories of the asset. Applicable to buildings.
        demand_config: dict
            A dictionary containing configuration options for the
            sample generation. Key options include:
            * 'SampleSize': The number of samples to generate.
            * 'PreserveRawOrder': Boolean indicating whether to
            preserve the order of the raw data. Defaults to False.
            * 'DemandCloning': Specifies if and how demand cloning
            should be applied. Can be a boolean or a detailed
            configuration.
        demand_data_source: string or dict
            If string, the demand_data_source is a file prefix
            (<prefix> in the following description) that identifies
            the following files: <prefix>_marginals.csv,
            <prefix>_empirical.csv, <prefix>_correlation.csv. If dict,
            the demand data source is a dictionary with the following
            optional keys: 'marginals', 'empirical', and
            'correlation'. The value under each key shall be a
            DataFrame.
        cmp_data_source: str or dict
            The source from where to load the component model data. If
            it's a string, it should be the prefix for three files:
            one for marginal distributions (`<prefix>_marginals.csv`),
            one for empirical data (`<prefix>_empirical.csv`), and one
            for correlation data (`<prefix>_correlation.csv`). If it's
            a dictionary, it should have keys 'marginals',
            'empirical', and 'correlation', with each key associated
            with a DataFrame containing the corresponding data.
        damage_data_paths: list of (string | DataFrame)
            List of paths to data or files with damage model
            information. Default XY datasets can be accessed as
            PelicunDefault/XY. Order matters. Parameters defined in
            prior elements in the list take precedence over the same
            parameters in subsequent data paths. I.e., place the
            Default datasets in the back.
        dmg_process: dict, optional
            Allows simulating damage processes, where damage to some
            component can alter the damage state of other components.
        scaling_specification: dict, optional
            A dictionary defining the shift in median.
            Example: {'CMP-1-1': '*1.2', 'CMP-1-2': '/1.4'}
            The keys are individual components that should be present
            in the `capacity_sample`.  The values should be strings
            containing an operation followed by the value formatted as
            a float.  The operation can be '+' for addition, '-' for
            subtraction, '*' for multiplication, and '/' for division.
        residual_drift_configuration: dict
            Dictionary containing the following keys-values:
            - params: dict
            A dictionary containing parameters required for the
            estimation method, such as 'yield_drift', which is the
            drift at which yielding is expected to occur.
            - method: str, optional
            The method used to estimate the RID values. Currently,
            only 'FEMA P58' is implemented. Defaults to 'FEMA P58'.
        collapse_fragility_configuration: dict
            Dictionary containing the following keys-values:
            - label: str
            Label to use to extend the MultiIndex of the demand
            sample.
            - value: float
            Values to add to the rows of the additional column.
            - unit: str
            Unit that corresponds to the additional column.
            - location: str, optional
            Optional location, defaults to `0`.
            - direction: str, optional
            Optional direction, defaults to `1`.
        block_batch_size: int
            Maximum number of components in each batch.

        """
        # TODO(JVM): when we build the API docs, ensure the above is
        # properly rendered.

        self.demand.load_model(demand_data_source)
        self.demand.generate_sample(demand_config)

        if residual_drift_configuration:
            self.demand.estimate_RID_and_adjust_sample(
                residual_drift_configuration['parameters'],
                residual_drift_configuration['method'],
            )

        if collapse_fragility_configuration:
            self.demand.expand_sample(
                collapse_fragility_configuration['label'],
                collapse_fragility_configuration['value'],
                collapse_fragility_configuration['unit'],
            )

        self.stories = num_stories
        self.asset.load_cmp_model(cmp_data_source)
        self.asset.generate_cmp_sample()

        self.damage.load_model_parameters(
            damage_data_paths, set(self.asset.list_unique_component_ids())
        )
        self.damage.calculate(dmg_process, block_batch_size, scaling_specification)

    def calculate_loss(
        self,
        decision_variables: tuple[str, ...],
        loss_model_data_paths: list[str | pd.DataFrame],
        loss_map_path: str | pd.DataFrame | None = None,
        loss_map_policy: str | None = None,
    ) -> None:
        """
        Calculate loss.

        Parameters
        ----------
        decision_variables: tuple
            Defines the decision variables to be included in the loss
            calculations. Defaults to those supported, but fewer can be
            used if desired. When fewer are used, the loss parameters for
            those not used will not be required.
        loss_model_data_paths: list of (string | DataFrame)
            List of paths to data or files with loss model
            information. Default XY datasets can be accessed as
            PelicunDefault/XY. Order matters. Parameters defined in
            prior elements in the list take precedence over the same
            parameters in subsequent data paths. I.e., place the
            Default datasets in the back.
        loss_map_path: str or pd.DataFrame or None
            Path to a csv file or DataFrame object that maps
            components IDs to their loss parameter definitions.
        loss_map_policy: str or None
            If None, does not modify the loss map.
            If set to `fill`, each component ID that is present in
            the asset model but not in the loss map is mapped to
            itself, but `excessiveRID` is excluded.
            If set to `fill_all`, each component ID that is present in
            the asset model but not in the loss map is mapped to
            itself without exceptions.

        """
        self.loss.decision_variables = decision_variables
        self.loss.add_loss_map(loss_map_path, loss_map_policy)
        self.loss.load_model_parameters(loss_model_data_paths)
        self.loss.calculate()

    def aggregate_loss(
        self,
        replacement_configuration: (
            tuple[uq.RandomVariableRegistry, dict[str, float]] | None
        ) = None,
        loss_combination: dict | None = None,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Aggregate losses.

        Parameters
        ----------
        replacement_configuration: Tuple, optional
            Tuple containing a RandomVariableRegistry and a
            dictionary. The RandomVariableRegistry is defining
            building replacement consequence RVs for the active
            decision variables. The dictionary defines exceedance
            thresholds. If the aggregated value for a decision
            variable (conditioned on no replacement) exceeds the
            threshold, then replacement is triggered. This can happen
            for multiple decision variables at the same
            realization. The consequence keyword `replacement` is
            reserved to represent exclusive triggering of the
            replacement consequences, and other consequences are
            ignored for those realizations where replacement is
            triggered. When assigned to None, then `replacement` is
            still treated as an exclusive consequence (other
            consequences are set to zero when replacement is nonzero)
            but it is not being additionally triggered by the
            exceedance of any thresholds. The aggregated loss sample
            contains an additional column with information on whether
            replacement was already present or triggered by a
            threshold exceedance for each realization.
        loss_combination: dict, optional
            Dictionary defining how losses for specific components
            should be aggregated for a given decision variable. It has
            the following structure: {`dv`: {(`c1`, `c2`): `arr`,
            ...}, ...}, where `dv` is some decision variable, (`c1`,
            `c2`) is a tuple defining a component pair, `arr` is a NxN
            numpy array defining a combination table, and `...` means
            that more key-value pairs with the same schema can exist
            in the dictionaries.  The loss sample is expected to
            contain columns that include both `c1` and `c2` listed as
            the component. The combination is applied to all pairs of
            columns where the components are `c1` and `c2`, and all of
            the rest of the multiindex levels match (`loc`, `dir`,
            `uid`). This means, for example, that when combining wind
            and flood losses, the asset model should contain both a
            wind and a flood component defined at the same
            location-direction.  `arr` can also be an M-dimensional
            numpy array where each dimension has length N (NxNx...xN).
            This structure allows for the loss combination of M
            components.  In this case the (`c1`, `c2`) tuple should
            contain M elements instead of two.

        Notes
        -----
        Regardless of the value of the arguments, this method does not
        alter the state of the loss model, i.e., it does not modify
        the values of the `.sample` attributes.

        Returns
        -------
        tuple
            Dataframe with the aggregated loss of each realization,
            and another boolean dataframe with information on which DV
            thresholds were exceeded in each realization, triggering
            replacement. If no thresholds are specified it only
            contains False values.

        """
        output = self.loss.aggregate_losses(
            replacement_configuration, loss_combination, future=True
        )
        assert isinstance(output, tuple)
        return output


class DLCalculationAssessment(AssessmentBase):
    """Base class for the assessment objects used in `DL_calculation.py`."""

    __slots__: list[str] = []

    def calculate_demand(  # noqa: C901
        self,
        demand_path: Path,
        collapse_limits: dict[str, float] | None,
        length_unit: str | None,
        demand_calibration: dict | None,
        sample_size: int,
        demand_cloning: dict | None,
        residual_drift_inference: dict | None,
        *,
        coupled_demands: bool,
    ) -> None:
        """
        Calculate demands.

        Parameters
        ----------
        demand_path: str
            Path to the demand data file.
        collapse_limits: dict[str, float] or None
            Optional dictionary with demand types and their respective
            collapse limits.
        length_unit : str, optional
            Unit of length to be used to add units to the demand data
            if needed.
        demand_calibration: dict or None
            Calibration data for the demand model.
        sample_size: int
            Number of realizations.
        coupled_demands: bool
            Whether to preserve the raw order of the demands.
        demand_cloning: dict or None
            Demand cloning configuration.
        residual_drift_inference: dict or None
            Information for residual drift inference.

        Raises
        ------
        ValueError
            When an unknown residual drift method is specified.

        """
        idx = pd.IndexSlice
        raw_demands = pd.read_csv(demand_path, index_col=0)

        # remove excessive demands that are considered collapses, if needed
        if collapse_limits:
            raw_demands_m = base.convert_to_MultiIndex(raw_demands, axis=1)
            assert isinstance(raw_demands_m, pd.DataFrame)
            raw_demands = raw_demands_m

            if 'Units' in raw_demands.index:
                raw_units = raw_demands.loc['Units', :]
                raw_demands = raw_demands.drop('Units', axis=0).astype(float)

            else:
                raw_units = None

            dem_to_drop = np.full(raw_demands.shape[0], fill_value=False)

            for dem_type, limit in collapse_limits.items():
                assert isinstance(dem_type, str)
                assert isinstance(limit, (str, float))
                nlevels_with_event_id = 4
                if raw_demands.columns.nlevels == nlevels_with_event_id:
                    dem_to_drop += raw_demands.loc[
                        :,  # type: ignore
                        idx[:, dem_type, :, :],
                    ].max(axis=1) > float(limit)

                else:
                    dem_to_drop += raw_demands.loc[
                        :,  # type: ignore
                        idx[dem_type, :, :],
                    ].max(axis=1) > float(limit)

            raw_demands = raw_demands.loc[~dem_to_drop, :]

            if isinstance(raw_units, pd.Series):
                raw_demands = pd.concat(
                    [raw_demands, raw_units.to_frame().T], axis=0
                )

            self.log.msg(
                f'{np.sum(dem_to_drop)} realizations removed from the demand '
                f'input because they exceed the collapse limit. The remaining '
                f'sample size: {raw_demands.shape[0]}'
            )

        # add units to the demand data if needed
        if 'Units' not in raw_demands.index:
            if length_unit is None:
                msg = 'A length unit is required to infer demand units.'
                raise ValueError(msg)
            demands = _add_units(raw_demands, length_unit)

        else:
            demands = raw_demands

        # load the available demand sample
        self.demand.load_sample(demands)

        # get the calibration information
        if demand_calibration:
            # then use it to calibrate the demand model
            self.demand.calibrate_model(demand_calibration)

        else:
            # if no calibration is requested,
            # set all demands to use empirical distribution
            self.demand.calibrate_model({'ALL': {'DistributionFamily': 'empirical'}})

        # and generate a new demand sample
        self.demand.generate_sample(
            {
                'SampleSize': sample_size,
                'PreserveRawOrder': coupled_demands,
                'DemandCloning': demand_cloning,
            }
        )

        # get the generated demand sample
        demand_sample_tuple = self.demand.save_sample(save_units=True)
        assert demand_sample_tuple is not None
        demand_sample, demand_units = demand_sample_tuple
        assert isinstance(demand_sample, pd.DataFrame)
        assert isinstance(demand_units, pd.Series)

        demand_sample = pd.concat([demand_sample, demand_units.to_frame().T])

        # get residual drift estimates, if needed
        if residual_drift_inference:
            # `method` is guaranteed to exist because it is confirmed when
            # parsing the configuration file.
            rid_inference_method = residual_drift_inference.pop('method')

            if rid_inference_method == 'FEMA P-58':
                rid_list: list[pd.DataFrame] = []
                pid = demand_sample['PID'].copy()
                pid = pid.drop('Units')
                pid = pid.astype(float)

                for direction, delta_yield in residual_drift_inference.items():
                    pids = pid.loc[:, idx[:, direction]]  # type: ignore
                    assert isinstance(pids, pd.DataFrame)
                    rid = self.demand.estimate_RID(
                        pids,
                        {'yield_drift': float(delta_yield)},
                    )

                    rid_list.append(rid)

                rid = pd.concat(rid_list, axis=1)
                rid_units = pd.Series(
                    ['unitless'] * rid.shape[1],
                    index=rid.columns,
                    name='Units',
                )
                rid_sample = pd.concat([rid, rid_units.to_frame().T])
                demand_sample = pd.concat([demand_sample, rid_sample], axis=1)

            else:
                msg = (
                    f'Unknown residual drift inference method: '
                    f'`{rid_inference_method}`.'
                )
                raise ValueError(msg)

        # add a constant one demand
        demand_sample['ONE', '0', '1'] = np.ones(demand_sample.shape[0])
        demand_sample.loc['Units', ('ONE', '0', '1')] = 'unitless'

        self.demand.load_sample(base.convert_to_SimpleIndex(demand_sample, axis=1))

    def calculate_asset(
        self,
        num_stories: int,
        component_assignment_file: str | None,
        collapse_fragility_demand_type: str | None,
        component_sample_file: str | None,
        *,
        add_irreparable_damage_columns: bool,
    ) -> None:
        """
        Generate the asset model sample.

        Parameters
        ----------
        num_stories: int
            Number of stories.
        component_assignment_file: str or None
            Path to a component assignment file.
        collapse_fragility_demand_type: str or None
            Optional demand type for the collapse fragility.
        add_irreparable_damage_columns: bool
            Whether to add columns for irreparable damage.
        component_sample_file: str or None
            Optional path to an existing component sample file.

        Raises
        ------
        ValueError
            With invalid combinations of arguments.

        """
        # retrieve the demand sample
        demand_sample = self.demand.save_sample()
        assert isinstance(demand_sample, pd.DataFrame)

        # set the number of stories
        if num_stories:
            self.stories = num_stories

        # We either accept a `component_assignment_file` or a
        # `component_sample_file`, not both.
        if (
            component_assignment_file is not None
            and component_sample_file is not None
        ):
            msg = (
                'Both `component_assignment_file` and '
                '`component_sample_file` are provided. '
                'Please provide only one.'
            )
            raise ValueError(msg)

        # load a component model and generate a sample
        if component_assignment_file is not None:
            cmp_marginals = pd.read_csv(
                component_assignment_file,
                index_col=0,
                encoding_errors='replace',
            )

            dem_types = demand_sample.columns.unique(level=0)

            # add component(s) to support collapse calculation
            if collapse_fragility_demand_type is not None:
                if not collapse_fragility_demand_type.startswith('SA'):
                    # we need story-specific collapse assessment
                    # (otherwise we have a global demand and evaluate
                    # collapse directly, so this code should be skipped)

                    if collapse_fragility_demand_type in dem_types:
                        # excessive coll_DEM is added on every floor
                        # to detect large RIDs
                        cmp_marginals.loc['excessive.coll.DEM', 'Units'] = 'ea'

                        locs = demand_sample[
                            collapse_fragility_demand_type  # type: ignore
                        ].columns.unique(level=0)
                        cmp_marginals.loc['excessive.coll.DEM', 'Location'] = (
                            ','.join(locs)
                        )

                        dirs = demand_sample[
                            collapse_fragility_demand_type  # type: ignore
                        ].columns.unique(level=1)
                        cmp_marginals.loc['excessive.coll.DEM', 'Direction'] = (
                            ','.join(dirs)
                        )

                        cmp_marginals.loc['excessive.coll.DEM', 'Theta_0'] = 1.0

                    else:
                        self.log.msg(
                            f'WARNING: No {collapse_fragility_demand_type} '
                            f'among available demands. Collapse cannot '
                            f'be evaluated.'
                        )

            # always add a component to support basic collapse calculation
            cmp_marginals.loc['collapse', 'Units'] = 'ea'
            cmp_marginals.loc['collapse', 'Location'] = 0
            cmp_marginals.loc['collapse', 'Direction'] = 1
            cmp_marginals.loc['collapse', 'Theta_0'] = 1.0

            # add components to support irreparable damage calculation
            if add_irreparable_damage_columns:
                if 'RID' in dem_types:
                    # excessive RID is added on every floor to detect large RIDs
                    cmp_marginals.loc['excessiveRID', 'Units'] = 'ea'

                    locs = demand_sample['RID'].columns.unique(level=0)
                    cmp_marginals.loc['excessiveRID', 'Location'] = ','.join(locs)

                    dirs = demand_sample['RID'].columns.unique(level=1)
                    cmp_marginals.loc['excessiveRID', 'Direction'] = ','.join(dirs)

                    cmp_marginals.loc['excessiveRID', 'Theta_0'] = 1.0

                    # irreparable is a global component to recognize is any of the
                    # excessive RIDs were triggered
                    cmp_marginals.loc['irreparable', 'Units'] = 'ea'
                    cmp_marginals.loc['irreparable', 'Location'] = 0
                    cmp_marginals.loc['irreparable', 'Direction'] = 1
                    cmp_marginals.loc['irreparable', 'Theta_0'] = 1.0

                else:
                    self.log.msg(
                        'WARNING: No residual interstory drift ratio among '
                        'available demands. Irreparable damage cannot be '
                        'evaluated.'
                    )

            # load component model
            self.asset.load_cmp_model({'marginals': cmp_marginals})

            # generate component quantity sample
            self.asset.generate_cmp_sample()

        # if requested, load the quantity sample from a file
        if component_sample_file is not None:
            self.asset.load_cmp_sample(component_sample_file)

    def calculate_damage(  # noqa: C901
        self,
        length_unit: str | None,
        component_database: str,
        component_database_path: str | None = None,
        collapse_fragility: dict | None = None,
        irreparable_damage: dict | None = None,
        damage_process_approach: str | None = None,
        damage_process_file_path: str | None = None,
        custom_model_dir: str | None = None,
        scaling_specification: dict | None = None,
        *,
        is_for_water_network_assessment: bool = False,
    ) -> None:
        """
        Calculate damage.

        Parameters
        ----------
        length_unit : str, optional
            Unit of length to be used to add units to the demand data
            if needed.
        component_database: str
            Name of the component database.
        component_database_path: str or None
            Optional path to a component database file.
        collapse_fragility: dict or None
            Collapse fragility information.
        irreparable_damage: dict or None
            Information for irreparable damage.
        damage_process_approach: str or None
            Approach for the damage process.
        damage_process_file_path: str or None
            Optional path to a damage process file.
        custom_model_dir: str or None
            Optional directory for custom models.
        scaling_specification: dict, optional
            A dictionary defining the shift in median.
            Example: {'CMP-1-1': '*1.2', 'CMP-1-2': '/1.4'}
            The keys are individual components that should be present
            in the `capacity_sample`.  The values should be strings
            containing an operation followed by the value formatted as
            a float.  The operation can be '+' for addition, '-' for
            subtraction, '*' for multiplication, and '/' for division.
        is_for_water_network_assessment: bool
            Whether the assessment is for a water network.

        Raises
        ------
        ValueError
            With invalid combinations of arguments.

        """
        # load the fragility information
        component_db = []

        if not pd.isna(component_database):
            for method_name in [
                cdb.strip() for cdb in component_database.split(',')
            ]:
                if method_name == 'None':
                    continue

                # <backwards compatibility>
                if method_name.endswith(('csv', 'CSV')):
                    component_db_path = file_io.substitute_default_path(
                        [f'PelicunDefault/{method_name}'], log=self.log
                    )[0]
                else:
                    component_db_path = file_io.substitute_default_path(
                        [f'PelicunDefault/{method_name}/fragility.csv'], log=self.log
                    )[0]
                assert isinstance(component_db_path, str)

                if Path(component_db_path).is_file():
                    component_db.append(component_db_path)

        if component_database_path is not None:
            if 'CustomDLDataFolder' in component_database_path:
                if custom_model_dir is None:
                    msg = (
                        '`custom_model_dir` needs to be specified '
                        'when `component_database_path` includes CustomDLDataFolder.'
                    )
                    raise ValueError(msg)

                component_database_path = component_database_path.replace(
                    'CustomDLDataFolder', custom_model_dir
                )

            component_db += [component_database_path]

        component_db = component_db[::-1]

        # prepare additional fragility data

        # get the database header from the default P58 db
        p58_data = self.get_default_data('FEMA P-58', 'fragility')

        adf = pd.DataFrame(columns=p58_data.columns)

        if collapse_fragility:
            assert self.asset.cmp_marginal_params is not None

            if (
                'excessive.coll.DEM'
                in self.asset.cmp_marginal_params.index.get_level_values('cmp')
            ):
                # if there is story-specific evaluation
                coll_cmp_name = 'excessive.coll.DEM'
            else:
                # otherwise, for global collapse evaluation
                coll_cmp_name = 'collapse'

            adf.loc[coll_cmp_name, ('Demand', 'Directional')] = 1
            adf.loc[coll_cmp_name, ('Demand', 'Offset')] = 0

            coll_dem = collapse_fragility['DemandType']

            if '_' in coll_dem:
                coll_dem, coll_dem_spec = coll_dem.split('_')
            else:
                coll_dem_spec = None

            coll_dem_name = None
            for demand_name, demand_short in EDP_to_demand_type.items():
                if demand_short == coll_dem:
                    coll_dem_name = demand_name
                    break

            if coll_dem_name is None:
                msg = (
                    'A valid demand type acronym was not provided in'
                    'the configuration file. Please ensure the'
                    "'DemandType' field in the collapse fragility"
                    'section contains one of the recognized acronyms'
                    "(e.g., 'SA', 'PFA', 'PGA'). Refer to the"
                    "configuration file's 'collapse_fragility'"
                    'section.'
                )
                raise ValueError(msg)

            if coll_dem_spec is None:
                adf.loc[coll_cmp_name, ('Demand', 'Type')] = coll_dem_name

            else:
                adf.loc[coll_cmp_name, ('Demand', 'Type')] = (
                    f'{coll_dem_name}|{coll_dem_spec}'
                )

            if length_unit is None:
                msg = 'A length unit is required.'
                raise ValueError(msg)
            coll_dem_unit = _add_units(
                pd.DataFrame(
                    columns=[
                        f'{coll_dem}-1-1',
                    ]
                ),
                length_unit,
            ).iloc[0, 0]

            adf.loc[coll_cmp_name, ('Demand', 'Unit')] = coll_dem_unit
            adf.loc[coll_cmp_name, ('LS1', 'Family')] = collapse_fragility[
                'CapacityDistribution'
            ]
            adf.loc[coll_cmp_name, ('LS1', 'Theta_0')] = collapse_fragility[
                'CapacityMedian'
            ]
            adf.loc[coll_cmp_name, ('LS1', 'Theta_1')] = collapse_fragility[
                'Theta_1'
            ]
            adf.loc[coll_cmp_name, 'Incomplete'] = 0

            if coll_cmp_name != 'collapse':
                # for story-specific evaluation, we need to add a placeholder
                # fragility that will never trigger, but helps us aggregate
                # results in the end
                adf.loc['collapse', ('Demand', 'Directional')] = 1
                adf.loc['collapse', ('Demand', 'Offset')] = 0
                adf.loc['collapse', ('Demand', 'Type')] = 'One'
                adf.loc['collapse', ('Demand', 'Unit')] = 'unitless'
                adf.loc['collapse', ('LS1', 'Theta_0')] = 1e10
                adf.loc['collapse', 'Incomplete'] = 0

        elif not is_for_water_network_assessment:
            # add a placeholder collapse fragility that will never trigger
            # collapse, but allow damage processes to work with collapse

            adf.loc['collapse', ('Demand', 'Directional')] = 1
            adf.loc['collapse', ('Demand', 'Offset')] = 0
            adf.loc['collapse', ('Demand', 'Type')] = 'One'
            adf.loc['collapse', ('Demand', 'Unit')] = 'unitless'
            adf.loc['collapse', ('LS1', 'Theta_0')] = 1e10
            adf.loc['collapse', 'Incomplete'] = 0

        if irreparable_damage:
            # add excessive RID fragility according to settings provided in the
            # input file
            adf.loc['excessiveRID', ('Demand', 'Directional')] = 1
            adf.loc['excessiveRID', ('Demand', 'Offset')] = 0
            adf.loc['excessiveRID', ('Demand', 'Type')] = (
                'Residual Interstory Drift Ratio'
            )

            adf.loc['excessiveRID', ('Demand', 'Unit')] = 'unitless'
            adf.loc['excessiveRID', ('LS1', 'Theta_0')] = irreparable_damage[
                'DriftCapacityMedian'
            ]
            adf.loc['excessiveRID', ('LS1', 'Family')] = 'lognormal'
            adf.loc['excessiveRID', ('LS1', 'Theta_1')] = irreparable_damage[
                'DriftCapacityLogStd'
            ]

            adf.loc['excessiveRID', 'Incomplete'] = 0

            # add a placeholder irreparable fragility that will never trigger
            # damage, but allow damage processes to aggregate excessiveRID here
            adf.loc['irreparable', ('Demand', 'Directional')] = 1
            adf.loc['irreparable', ('Demand', 'Offset')] = 0
            adf.loc['irreparable', ('Demand', 'Type')] = 'One'
            adf.loc['irreparable', ('Demand', 'Unit')] = 'unitless'
            adf.loc['irreparable', ('LS1', 'Theta_0')] = 1e10
            adf.loc['irreparable', 'Incomplete'] = 0

        # TODO(AZ): we can improve this by creating a water
        # network-specific assessment class
        if is_for_water_network_assessment:
            # add a placeholder aggregate fragility that will never trigger
            # damage, but allow damage processes to aggregate the
            # various pipeline damages
            adf.loc['aggregate', ('Demand', 'Directional')] = 1
            adf.loc['aggregate', ('Demand', 'Offset')] = 0
            adf.loc['aggregate', ('Demand', 'Type')] = 'Peak Ground Velocity'
            adf.loc['aggregate', ('Demand', 'Unit')] = 'mps'
            adf.loc['aggregate', ('LS1', 'Theta_0')] = 1e10
            adf.loc['aggregate', ('LS2', 'Theta_0')] = 1e10
            adf.loc['aggregate', 'Incomplete'] = 0

        self.damage.load_model_parameters(
            [*component_db, adf],
            set(self.asset.list_unique_component_ids()),
        )

        # load the damage process if needed
        dmg_process = None
        if damage_process_approach is not None:  # noqa: PLR1702
            if damage_process_approach in default_damage_processes:
                dmg_process = default_damage_processes[damage_process_approach]

                # For Hazus Earthquake, we need to specify the component ids
                if damage_process_approach in {
                    'Hazus Earthquake',
                    'Hazus Earthquake - Buildings',
                    'Hazus Earthquake - Lifeline Facilities',
                }:
                    cmp_sample = self.asset.save_cmp_sample()
                    assert isinstance(cmp_sample, pd.DataFrame)

                    cmp_list = cmp_sample.columns.unique(level=0)

                    cmp_map = {'STR': '', 'LF': '', 'NSA': ''}

                    for cmp in cmp_list:
                        for cmp_type in cmp_map:
                            if cmp_type + '.' in cmp:
                                cmp_map[cmp_type] = cmp

                    new_dmg_process = dmg_process.copy()
                    for source_cmp, action in dmg_process.items():
                        # first, look at the source component id
                        new_source = None
                        for cmp_type, cmp_id in cmp_map.items():
                            if (cmp_type in source_cmp) and (cmp_id != ''):  # noqa: PLC1901
                                new_source = source_cmp.replace(cmp_type, cmp_id)
                                break

                        if new_source is not None:
                            new_dmg_process[new_source] = action
                            del new_dmg_process[source_cmp]
                        else:
                            new_source = source_cmp

                        # then, look at the target component ids
                        for ds_i, target_vals in action.items():
                            if isinstance(target_vals, str):
                                for cmp_type, cmp_id in cmp_map.items():
                                    if (cmp_type in target_vals) and (cmp_id != ''):  # noqa: PLC1901
                                        target_vals = target_vals.replace(  # noqa: PLW2901
                                            cmp_type, cmp_id
                                        )

                                new_target_vals = target_vals

                            else:
                                # we assume that target_vals is a list of str
                                new_target_vals = []

                                for target_val in target_vals:
                                    for cmp_type, cmp_id in cmp_map.items():
                                        if (cmp_type in target_val) and (
                                            cmp_id != ''  # noqa: PLC1901
                                        ):
                                            target_val = target_val.replace(  # noqa: PLW2901
                                                cmp_type, cmp_id
                                            )

                                    new_target_vals.append(target_val)

                            new_dmg_process[new_source][ds_i] = new_target_vals

                    dmg_process = new_dmg_process

                # Remove components not present in the asset model
                # from the source components of the damage process.
                asset_components = set(self.asset.list_unique_component_ids())
                filtered_dmg_process = {}
                for key in dmg_process:
                    component = key.split('_')[1]
                    if component in asset_components:
                        filtered_dmg_process[key] = dmg_process[key]
                dmg_process = filtered_dmg_process

            elif damage_process_approach == 'User Defined':
                if damage_process_file_path is None:
                    msg = (
                        'When `damage_process_approach` is set to '
                        '`User Defined`, a `damage_process_file_path` '
                        'needs to be provided.'
                    )
                    raise ValueError(msg)

                # load the damage process from a file
                with Path(damage_process_file_path).open(encoding='utf-8') as f:
                    dmg_process = json.load(f)

            elif damage_process_approach == 'None':
                # no damage process applied for the calculation
                dmg_process = None

            else:
                self.log.msg(
                    f'Prescribed Damage Process not recognized: '
                    f'`{damage_process_approach}`.'
                )

        # calculate damages
        self.damage.calculate(
            dmg_process=dmg_process,
            scaling_specification=scaling_specification,
        )

    def calculate_loss(
        self,
        loss_map_approach: str,
        occupancy_type: str,
        consequence_database: str,
        consequence_database_path: str | None = None,
        custom_model_dir: str | None = None,
        damage_process_approach: str = 'User Defined',
        replacement_cost_parameters: dict[str, float | str] | None = None,
        replacement_time_parameters: dict[str, float | str] | None = None,
        replacement_carbon_parameters: dict[str, float | str] | None = None,
        replacement_energy_parameters: dict[str, float | str] | None = None,
        loss_map_path: str | None = None,
        decision_variables: tuple[str, ...] | None = None,
        replacement_configuration: (
            tuple[uq.RandomVariableRegistry, dict[str, float]] | None
        ) = None,
        loss_combination_method: str | None = None,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Calculate losses.

        Parameters
        ----------
        loss_map_approach: str
            Approach for the loss map generation. Can be either
            `User Defined` or `Automatic`.
        occupancy_type: str
            Occupancy type.
        consequence_database: str
            Name of the consequence database.
        consequence_database_path: str or None
            Optional path to a consequence database file.
        custom_model_dir: str or None
            Optional directory for custom models.
        damage_process_approach: str
            Damage process approach. Defaults to `User Defined`.
        replacement_cost_parameters: dict or None
            Parameters for replacement cost.
        replacement_time_parameters: dict or None
            Parameters for replacement time.
        replacement_carbon_parameters: dict or None
            Parameters for replacement carbon.
        replacement_energy_parameters: dict or None
            Parameters for replacement energy.
        loss_map_path: str or None
            Optional path to a loss map file.
        decision_variables: tuple[str] or None
            Optional decision variables for the assessment.
        replacement_configuration: tuple or None
            Loss thresholds of replacement consequences.
        loss_combination_method: str, optional
            String defining the method to use for combining losses for
            components that represent different demands.

        Returns
        -------
        tuple
            Dataframe with the aggregated loss of each realization,
            and another boolean dataframe with information on which DV
            thresholds were exceeded in each realization, triggering
            replacement. If no thresholds are specified it only
            contains False values.

        Raises
        ------
        ValueError
            When an invalid loss map approach is specified.

        """
        conseq_df, consequence_db = self.load_consequence_info(
            consequence_database,
            consequence_database_path,
            custom_model_dir,
        )

        # remove duplicates from conseq_df
        conseq_df = conseq_df.loc[conseq_df.index.unique(), :]

        # add the replacement consequence to the data
        adf = pd.DataFrame(
            columns=conseq_df.columns,
            index=pd.MultiIndex.from_tuples(
                [
                    ('replacement', 'Cost'),
                    ('replacement', 'Time'),
                    ('replacement', 'Carbon'),
                    ('replacement', 'Energy'),
                ]
            ),
        )

        _loss__add_replacement_cost(
            adf,
            damage_process_approach,
            unit=get(replacement_cost_parameters, 'Unit'),
            median=get(replacement_cost_parameters, 'Median'),
            distribution=get(replacement_cost_parameters, 'Distribution'),
            theta_1=get(replacement_cost_parameters, 'Theta_1'),
        )

        _loss__add_replacement_time(
            adf,
            damage_process_approach,
            conseq_df,
            occupancy_type=occupancy_type,
            unit=get(replacement_time_parameters, 'Unit'),
            median=get(replacement_time_parameters, 'Median'),
            distribution=get(replacement_time_parameters, 'Distribution'),
            theta_1=get(replacement_time_parameters, 'Theta_1'),
        )

        _loss__add_replacement_carbon(
            adf,
            damage_process_approach,
            unit=get(replacement_carbon_parameters, 'Unit'),
            median=get(replacement_carbon_parameters, 'Median'),
            distribution=get(replacement_carbon_parameters, 'Distribution'),
            theta_1=get(replacement_carbon_parameters, 'Theta_1'),
        )

        _loss__add_replacement_energy(
            adf,
            damage_process_approach,
            unit=get(replacement_energy_parameters, 'Unit'),
            median=get(replacement_energy_parameters, 'Median'),
            distribution=get(replacement_energy_parameters, 'Distribution'),
            theta_1=get(replacement_energy_parameters, 'Theta_1'),
        )

        # prepare the loss map
        loss_map = None
        if loss_map_approach == 'Automatic':
            # get the damage sample
            loss_map = _loss__map_auto(
                self, conseq_df, damage_process_approach, occupancy_type
            )

        elif loss_map_approach == 'User Defined':
            assert custom_model_dir is not None
            loss_map = _loss__map_user(custom_model_dir, loss_map_path)

        else:
            msg = f'Invalid MapApproach value: `{loss_map_approach}`.'
            raise ValueError(msg)

        # prepare additional loss map entries, if needed
        if 'DMG-collapse' not in loss_map.index:
            loss_map.loc['collapse', 'Repair'] = 'replacement'
            loss_map.loc['irreparable', 'Repair'] = 'replacement'

        if decision_variables:
            self.loss.decision_variables = decision_variables

        self.loss.add_loss_map(loss_map, loss_map_policy=None)
        self.loss.load_model_parameters([*consequence_db, adf])

        self.loss.calculate()

        if loss_combination_method is None:
            loss_combination = None

        elif loss_combination_method == 'Hazus Hurricane':
            # assemble the combination dict for wind and storm surge
            # open the base combination matrix
            file_path = file_io.substitute_default_path(
                ['PelicunDefault/Hazus Hurricane Wind/combine_wind_flood.csv'],
                log=self.log,
            )[0]
            assert isinstance(file_path, str)
            combination_array = pd.read_csv(
                file_path,
                index_col=None,
                header=None,
            ).to_numpy()

            # get the component names
            # assume that the first and second component in the loss map
            # are the wind and flood components, respectively
            wind_comp, flood_comp = loss_map.index.to_numpy()[[0, 1]]

            loss_combination = {
                'Cost': {
                    (wind_comp, flood_comp): combination_array,
                },
            }

        else:
            msg = f'Invalid loss combination method: `{loss_combination_method}`.'
            raise ValueError(msg)

        df_agg, exceedance_bool_df = self.loss.aggregate_losses(
            replacement_configuration, loss_combination, future=True
        )
        assert isinstance(df_agg, pd.DataFrame)
        assert isinstance(exceedance_bool_df, pd.DataFrame)
        return df_agg, exceedance_bool_df

    def load_consequence_info(
        self,
        consequence_database: str,
        consequence_database_path: str | None = None,
        custom_model_dir: str | None = None,
    ) -> tuple[pd.DataFrame, list[str]]:
        """
        Load consequence information for the assessment.

        Parameters
        ----------
        consequence_database: str
            Name of the consequence database.
        consequence_database_path: str or None
            Optional path to a consequence database file.
        custom_model_dir: str or None
            Optional directory for custom models.

        Returns
        -------
        tuple[pd.DataFrame, list[str]]
            A tuple containing:
            - A DataFrame with the consequence data.
            - A list of paths to the consequence databases used.

        Raises
        ------
        ValueError
            With invalid combinations of arguments.

        """
        # load the consequence information
        consequence_db = []
        conseq_df = pd.DataFrame()

        if not pd.isna(consequence_database):
            for method_name in [
                cdb.strip() for cdb in consequence_database.split(',')
            ]:
                if method_name == 'None':
                    continue

                # <backwards compatibility>
                if method_name.endswith(('csv', 'CSV')):
                    consequence_db_path = file_io.substitute_default_path(
                        [f'PelicunDefault/{method_name}'], log=self.log
                    )[0]
                else:
                    consequence_db_path = file_io.substitute_default_path(
                        [f'PelicunDefault/{method_name}/consequence_repair.csv'],
                        log=self.log,
                    )[0]
                assert isinstance(consequence_db_path, str)

                if Path(consequence_db_path).is_file():
                    consequence_db.append(consequence_db_path)

                    conseq_df = pd.concat(
                        [
                            conseq_df,
                            self.get_default_data(method_name, 'consequence_repair'),
                        ]
                    )
                    assert isinstance(conseq_df, pd.DataFrame)

                else:
                    # try loading loss functions instead
                    loss_db_path = file_io.substitute_default_path(
                        [f'PelicunDefault/{method_name}/loss_repair.csv'],
                        log=self.log,
                    )[0]

                    if Path(loss_db_path).is_file():
                        consequence_db.append(loss_db_path)

                        conseq_df = pd.concat(
                            [
                                conseq_df,
                                self.get_default_data(method_name, 'loss_repair'),
                            ]
                        )
                        assert isinstance(conseq_df, pd.DataFrame)

        if consequence_database_path is not None:
            if 'CustomDLDataFolder' in consequence_database_path:
                if custom_model_dir is None:
                    msg = (
                        'When `consequence_database_path` includes CustomDLDataFolder, '
                        '`custom_model_dir` needs to be specified as well.'
                    )
                    raise ValueError(msg)

                consequence_database_path = consequence_database_path.replace(
                    'CustomDLDataFolder', custom_model_dir
                )

            consequence_db += [consequence_database_path]

            extra_conseq_df = file_io.load_data(
                consequence_database_path,
                unit_conversion_factors=None,
                orientation=1,
                reindex=False,
                log=self.log,
            )
            assert isinstance(extra_conseq_df, pd.DataFrame)

            conseq_df = pd.concat([conseq_df, extra_conseq_df])
            assert isinstance(conseq_df, pd.DataFrame)

        consequence_db = consequence_db[::-1]

        return conseq_df, consequence_db


def _add_units(raw_demands: pd.DataFrame, length_unit: str) -> pd.DataFrame:
    """
    Add units to demand columns in a DataFrame.

    Parameters
    ----------
    raw_demands: pd.DataFrame
        The raw demand data to which units will be added.
    length_unit: str
        The unit of length to be used (e.g., 'in' for inches).

    Returns
    -------
    pd.DataFrame
        The DataFrame with units added to the appropriate demand columns.

    """
    demands = raw_demands.T

    demands.insert(0, 'Units', np.nan)

    if length_unit == 'in':
        length_unit = 'inch'

    demands = pd.DataFrame(
        base.convert_to_MultiIndex(demands, axis=0).sort_index(axis=0).T
    )

    nlevels_with_event_id = 4
    dem_level = 1 if demands.columns.nlevels == nlevels_with_event_id else 0

    # drop demands with no EDP type identified
    demands = demands.drop(
        demands.columns[demands.columns.get_level_values(dem_level) == ''],
        axis=1,
    )

    # assign units
    demand_cols = demands.columns.get_level_values(dem_level).to_list()

    # remove additional info from demand names
    demand_cols = [d.split('_')[0] for d in demand_cols]

    # acceleration
    acc_edps = ['PFA', 'PGA', 'SA']
    edp_mask = np.isin(demand_cols, acc_edps)

    if np.any(edp_mask):
        demands.iloc[0, edp_mask] = length_unit + 'ps2'  # type: ignore

    # speed
    speed_edps = ['PFV', 'PWS', 'PGV', 'SV']
    edp_mask = np.isin(demand_cols, speed_edps)

    if np.any(edp_mask):
        demands.iloc[0, edp_mask] = length_unit + 'ps'  # type: ignore

    # displacement
    disp_edps = ['PFD', 'PIH', 'SD', 'PGD']
    edp_mask = np.isin(demand_cols, disp_edps)

    if np.any(edp_mask):
        demands.iloc[0, edp_mask] = length_unit  # type: ignore

    # drift ratio
    rot_edps = ['PID', 'PRD', 'DWD', 'RDR', 'PMD', 'RID']
    edp_mask = np.isin(demand_cols, rot_edps)

    if np.any(edp_mask):
        demands.iloc[0, edp_mask] = 'unitless'  # type: ignore

    # convert back to simple header and return the DF
    return base.convert_to_SimpleIndex(demands, axis=1)


def _loss__add_replacement_energy(
    adf: pd.DataFrame,
    dl_method: str,
    unit: str | None = None,
    median: float | None = None,
    distribution: str | None = None,
    theta_1: float | None = None,
) -> None:
    """
    Add replacement energy information.

    Parameters
    ----------
    adf : pandas.DataFrame
        Dataframe containing loss information.
    DL_method : str
        Supported methods are 'FEMA P-58'.
    unit : str, optional
        Unit for the energy value (e.g., 'MJ'). Defaults to None.
    median : float, optional
        Median replacement energy. If provided, it defines the base
        replacement energy value. Defaults to None.
    distribution : str, optional
        Distribution family to model uncertainty around the median
        energy (e.g., 'lognormal'). Required if `median` is
        provided. Defaults to None.
    theta_1 : float, optional
        Distribution parameter (e.g., standard deviation). Required if
        `distribution` is provided. Defaults to None.

    Notes
    -----
    If `median` is not provided, a default value is assigned based on
    the `DL_method`. For 'FEMA P-58', the default replacement energy
    value is 0 MJ.  For other methods, this consequence is removed
    from the dataframe entirely.
    """
    ren = ('replacement', 'Energy')
    if median is not None:
        # TODO(JVM): in this case we need unit (add config parser check)

        adf.loc[ren, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[ren, ('DV', 'Unit')] = unit
        adf.loc[ren, ('DS1', 'Theta_0')] = median

        if distribution is not None:
            # TODO(JVM): in this case we need theta_1 (add config parser check)

            adf.loc[ren, ('DS1', 'Family')] = distribution
            adf.loc[ren, ('DS1', 'Theta_1')] = theta_1
    elif dl_method == 'FEMA P-58':
        adf.loc[ren, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[ren, ('DV', 'Unit')] = 'MJ'
        adf.loc[ren, ('DS1', 'Theta_0')] = 0

    else:
        # for everything else, remove this consequence
        adf = adf.drop(ren)


def _loss__add_replacement_carbon(
    adf: pd.DataFrame,
    damage_process_approach: str,
    unit: str | None = None,
    median: float | None = None,
    distribution: str | None = None,
    theta_1: float | None = None,
) -> None:
    """
    Add replacement carbon emission information.

    Parameters
    ----------
    adf : pandas.DataFrame
        Dataframe containing loss information.
    damage_process_approach : str
        Supported approaches include 'FEMA P-58'.
    unit : str, optional
        Unit for the carbon emission value (e.g., 'kg'). Defaults to
        None.
    median : float, optional
        Median replacement carbon emissions. If provided, it defines
        the base replacement carbon value. Defaults to None.
    distribution : str, optional
        Distribution family to model uncertainty around the median
        carbon emissions (e.g., 'lognormal'). Required if `median` is
        provided.  Defaults to None.
    theta_1 : float, optional
        Distribution parameter (e.g., standard deviation). Required if
        `distribution` is provided. Defaults to None.

    Notes
    -----
    If `median` is not provided, a default value is assigned based on
    the `damage_process_approach`. For 'FEMA P-58', the default
    replacement carbon emissions value is 0 kg. For other approaches,
    this consequence is removed from the dataframe entirely.
    """
    rcarb = ('replacement', 'Carbon')
    if median is not None:
        # TODO(JVM): in this case we need unit (add config parser check)

        adf.loc[rcarb, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rcarb, ('DV', 'Unit')] = unit
        adf.loc[rcarb, ('DS1', 'Theta_0')] = median

        if distribution is not None:
            # TODO(JVM): in this case we need theta_1 (add config parser check)

            adf.loc[rcarb, ('DS1', 'Family')] = distribution
            adf.loc[rcarb, ('DS1', 'Theta_1')] = theta_1
    elif damage_process_approach == 'FEMA P-58':
        adf.loc[rcarb, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rcarb, ('DV', 'Unit')] = 'kg'
        adf.loc[rcarb, ('DS1', 'Theta_0')] = 0

    else:
        # for everything else, remove this consequence
        adf = adf.drop(rcarb)


def _loss__add_replacement_time(
    adf: pd.DataFrame,
    damage_process_approach: str,
    conseq_df: pd.DataFrame,
    occupancy_type: str | None = None,
    unit: str | None = None,
    median: float | None = None,
    distribution: str | None = None,
    theta_1: float | None = None,
) -> None:
    """
    Add replacement time information.

    Parameters
    ----------
    adf : pandas.DataFrame
        Dataframe containing loss information.
    damage_process_approach : str
        Supported approaches are 'FEMA P-58', 'Hazus Earthquake -
        Buildings'.
    conseq_df : pandas.DataFrame
        Dataframe containing consequence data for different damage
        states.
    occupancy_type : str, optional
        Type of occupancy, used to look up replacement time in the
        consequence dataframe for Hazus Earthquake approach. Defaults
        to None.
    unit : str, optional
        Unit for the replacement time (e.g., 'day, 'worker_day').
        Defaults to None.
    median : float, optional
        Median replacement time or loss ratio. If provided, it defines
        the base replacement time. Defaults to None.
    distribution : str, optional
        Distribution family to model uncertainty around the median
        time (e.g., 'lognormal'). Required if `median` is
        provided. Defaults to None.
    theta_1 : float, optional
        Distribution parameter (e.g., standard deviation). Required if
        `distribution` is provided. Defaults to None.

    Notes
    -----
    If `median` is not provided, a default value is assigned based on
    the `damage_process_approach`. For 'FEMA P-58', the default
    replacement time is 0 worker_days. For 'Hazus Earthquake -
    Buildings', the replacement time is fetched from `conseq_df` for
    the provided `occupancy_type` and corresponds to the total loss
    (damage state 5, DS5). In other cases, a placeholder value of 1 is
    used.

    """
    rt = ('replacement', 'Time')
    if median is not None:
        # TODO(JVM): in this case we need unit (add config parser check)

        adf.loc[rt, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rt, ('DV', 'Unit')] = unit
        adf.loc[rt, ('DS1', 'Theta_0')] = median

        if distribution is not None:
            # TODO(JVM): in this case we need theta_1 (add config parser check)

            adf.loc[rt, ('DS1', 'Family')] = distribution
            adf.loc[rt, ('DS1', 'Theta_1')] = theta_1
    elif damage_process_approach == 'FEMA P-58':
        adf.loc[rt, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rt, ('DV', 'Unit')] = 'worker_day'
        adf.loc[rt, ('DS1', 'Theta_0')] = 0

    # for Hazus EQ, use 1.0 as a loss_ratio
    elif damage_process_approach == 'Hazus Earthquake - Buildings':
        adf.loc[rt, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rt, ('DV', 'Unit')] = 'day'

        # load the replacement time that corresponds to total loss
        adf.loc[rt, ('DS1', 'Theta_0')] = conseq_df.loc[
            (f'STR.{occupancy_type}', 'Time'), ('DS5', 'Theta_0')
        ]

    elif damage_process_approach == 'Hazus Earthquake - Lifeline Facilities':
        adf.loc[rt, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rt, ('DV', 'Unit')] = 'day'

        # load the replacement time that corresponds to total loss
        adf.loc[rt, ('DS1', 'Theta_0')] = conseq_df.loc[
            (f'LF.{occupancy_type}', 'Time'), ('DS5', 'Theta_0')
        ]

    # otherwise, use 1 (and expect to have it defined by the user)
    else:
        adf.loc[rt, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rt, ('DV', 'Unit')] = 'loss_ratio'
        adf.loc[rt, ('DS1', 'Theta_0')] = 1


def _loss__add_replacement_cost(
    adf: pd.DataFrame,
    dl_method: str,
    unit: str | None = None,
    median: float | None = None,
    distribution: str | None = None,
    theta_1: float | None = None,
) -> None:
    """
    Add replacement cost information.

    Parameters
    ----------
    adf : pandas.DataFrame
        Dataframe containing loss information.
    DL_method : str
        Supported methods are 'FEMA P-58', 'Hazus Earthquake', and
        'Hazus Hurricane'.
    unit : str, optional
        Unit for the replacement cost (e.g., 'USD_2011',
        'loss_ratio'). Defaults to None.
    median : float, optional
        Median replacement cost or loss ratio. If provided, it defines
        the base replacement cost. Defaults to None.
    distribution : str, optional
        Distribution family to model uncertainty around the median
        cost (e.g., 'lognormal'). Required if `median` is
        provided. Defaults to None.
    theta_1 : float, optional
        Distribution parameter (e.g., standard deviation). Required if
        `distribution` is provided. Defaults to None.
    """
    rc = ('replacement', 'Cost')
    if median is not None:
        # TODO(JVM): in this case we need unit (add config parser check)

        adf.loc[rc, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rc, ('DV', 'Unit')] = unit
        adf.loc[rc, ('DS1', 'Theta_0')] = median

        if distribution is not None:
            # TODO(JVM): in this case we need theta_1 (add config parser check)

            adf.loc[rc, ('DS1', 'Family')] = distribution
            adf.loc[rc, ('DS1', 'Theta_1')] = theta_1

    elif dl_method == 'FEMA P-58':
        adf.loc[rc, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rc, ('DV', 'Unit')] = 'USD_2011'
        adf.loc[rc, ('DS1', 'Theta_0')] = 0

    # for Hazus EQ and HU, use 1.0 as a loss_ratio
    elif dl_method in {'Hazus Earthquake', 'Hazus Hurricane'}:
        adf.loc[rc, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rc, ('DV', 'Unit')] = 'loss_ratio'

        # store the replacement cost that corresponds to total loss
        adf.loc[rc, ('DS1', 'Theta_0')] = 1.00

    # otherwise, use 1 (and expect to have it defined by the user)
    else:
        adf.loc[rc, ('Quantity', 'Unit')] = '1 EA'
        adf.loc[rc, ('DV', 'Unit')] = 'loss_ratio'
        adf.loc[rc, ('DS1', 'Theta_0')] = 1


def _loss__map_user(
    custom_model_dir: str, loss_map_path: str | None = None
) -> pd.DataFrame:
    """
    Load a user-defined loss map from a specified path.

    Parameters
    ----------
    custom_model_dir : str
        Directory containing custom models.
    loss_map_path : str, optional
        Path to the loss map file. The path can include a placeholder
        'CustomDLDataFolder' that will be replaced by
        `custom_model_dir`.  If not provided, raises a ValueError.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the loss map information.

    Raises
    ------
    ValueError
        If `loss_map_path` is not provided.

    """
    if loss_map_path is not None:
        loss_map_path = loss_map_path.replace('CustomDLDataFolder', custom_model_dir)

    else:
        msg = 'Missing loss map path.'
        raise ValueError(msg)

    return pd.read_csv(loss_map_path, index_col=0)


def _loss__map_auto(
    assessment: DLCalculationAssessment,
    conseq_df: pd.DataFrame,
    dl_method: str,
    occupancy_type: str | None = None,
) -> pd.DataFrame:
    """
    Automatically generate a loss map.

    Automatically generate a loss map based on the damage sample and
    the consequence database.

    Parameters
    ----------
    assessment : AssessmentBase
        The assessment object containing the damage model and sample.
    conseq_df : pandas.DataFrame
        DataFrame containing consequence data for different damage
        states.
    DL_method : str
        Damage loss method, which defines how the loss map is
        generated.  Supported methods are 'FEMA P-58', 'Hazus
        Earthquake', 'Hazus Hurricane', and 'Hazus Earthquake
        Transportation'.
    occupancy_type : str, optional
        Occupancy type, used to map damage components to the correct
        loss models in Hazus Earthquake methods. Defaults to None.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the automatically generated loss map,
        where the index corresponds to the damage components and the
        values indicate the associated loss models.

    Notes
    -----
    - For 'FEMA P-58' and 'Hazus Hurricane', the method assumes that
      fragility and consequence data have matching component IDs.
    - For 'Hazus Earthquake' and 'Hazus Earthquake Transportation',
      the method assumes that consequence archetypes are only
      differentiated by occupancy type.

    """
    identical_damage_and_conseqence_ids = True
    if dl_method.startswith('Hazus Earthquake'):
        identical_damage_and_conseqence_ids = False

    # get the component sample
    asset_sample = assessment.asset.save_cmp_sample()
    assert isinstance(asset_sample, pd.DataFrame)

    # get the damage sample
    # TODO(AZ): check why the damage sample was needed here
    # dmg_sample = assessment.damage.save_sample()
    # assert isinstance(dmg_sample, pd.DataFrame)

    # create a mapping for all components that are also in
    # the prescribed consequence database
    asset_cmps = asset_sample.columns.unique(level='cmp')
    loss_cmps = conseq_df.index.unique(level=0)

    drivers = []
    loss_models = []

    if identical_damage_and_conseqence_ids:
        # with these methods, we assume fragility and consequence data
        # have the same IDs

        for asset_cmp in asset_cmps:
            if asset_cmp == 'collapse':
                continue

            if asset_cmp in loss_cmps:
                drivers.append(asset_cmp)
                loss_models.append(asset_cmp)

    else:
        # Currently, we only get here with Hazus Earthquake
        # with Hazus Earthquake we assume that consequence
        # archetypes are only differentiated by occupancy type
        for asset_cmp in asset_cmps:
            if asset_cmp == 'collapse':
                continue

            cmp_class = asset_cmp.split('.')[0]
            if occupancy_type is not None:
                loss_cmp = f'{cmp_class}.{occupancy_type}'
            else:
                loss_cmp = cmp_class

            if loss_cmp in loss_cmps:
                drivers.append(asset_cmp)
                loss_models.append(loss_cmp)

    return pd.DataFrame(loss_models, columns=['Repair'], index=drivers)


class TimeBasedAssessment:
    """Time-based assessment."""



# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

import time

import numpy as np
import pandas as pd


def benchmark():
    # Create a large DataFrame
    df = pd.DataFrame(np.random.rand(1000000, 10), columns=list('ABCDEFGHIJ'))

    # Measure time for df.to_dict(orient='list')
    start_time = time.time()
    df.to_dict(orient='list')
    end_time = time.time()
    print(f'Time taken with to_dict(orient="list"): {end_time - start_time} seconds')

    # Measure time for dictionary comprehension
    start_time = time.time()
    {col: df[col].tolist() for col in df.columns}
    end_time = time.time()
    print(
        f'Time taken with dictionary comprehension: {end_time - start_time} seconds'
    )

    # Measure time for dictionary comprehension without to list
    start_time = time.time()
    {col: df[col] for col in df.columns}
    end_time = time.time()
    print(
        f'Time taken with dictionary comprehension '
        f'without to_list: {end_time - start_time} seconds'
    )

    # Measure time for .values
    start_time = time.time()
    df.values
    end_time = time.time()
    print(f'Time taken with .values: {end_time - start_time} seconds')

    # Measure time for using df.to_numpy()
    start_time = time.time()
    data_array = df.to_numpy()
    {col: data_array[:, i].tolist() for i, col in enumerate(df.columns)}
    end_time = time.time()
    print(f'Time taken with df.to_numpy(): {end_time - start_time} seconds')

    # Measure time for using df.to_dict()
    start_time = time.time()
    df.to_dict()
    end_time = time.time()
    print(f'Time taken with df.to_dict(): {end_time - start_time} seconds')


if __name__ == '__main__':
    benchmark()



#
# Copyright (c) 2025 Leland Stanford Junior University
# Copyright (c) 2025 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay

"""These are integration tests on the dlml module of pelicun."""

from __future__ import annotations

import subprocess  # noqa: S404
import tempfile
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, mock_open, patch

import pytest
import requests

from pelicun.tools import dlml


def test_end_to_end_workflow_with_mock() -> None:
    """Test end-to-end workflow with mock GitHub API."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses
        mock_release_response = MagicMock()
        mock_release_response.json.return_value = {'target_commitish': '1234567'}

        # Mock API responses and the data directory path
        with patch('requests.get', return_value=mock_release_response), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch(
            'pathlib.Path.open', mock_open(read_data='file1.txt\nfile2.txt')
        ), patch('pelicun.tools.dlml.tqdm'):
            # Download data files
            dlml.download_data_files(version='v1.0.0', use_cache=False)

            # Verify _download_file was called
            assert mock_download.call_count == 3  # model_files.txt + 2 files


def test_with_mock_github_api_responses() -> None:
    """Test with mock GitHub API responses."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses for different API calls
        def mock_get_side_effect(url: str, **kwargs: Any) -> MagicMock:  # noqa: ARG001, ANN401
            if 'releases/tags' in url:
                mock_response = MagicMock()
                mock_response.json.return_value = {'target_commitish': '1234567'}
                return mock_response
            if 'commits' in url:
                mock_response = MagicMock()
                mock_response.json.return_value = [{'sha': '1234567'}]
                return mock_response
            mock_response = MagicMock()
            mock_response.iter_content.return_value = [b'chunk1', b'chunk2']
            return mock_response

        # Mock API responses and the data directory path
        with patch('requests.get', side_effect=mock_get_side_effect), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch(
            'pathlib.Path.open', mock_open(read_data='file1.txt\nfile2.txt')
        ), patch('pelicun.tools.dlml.tqdm'):
            # Test with version
            dlml.download_data_files(version='v1.0.0', use_cache=False)
            assert mock_download.call_count == 3  # model_files.txt + 2 files
            mock_download.reset_mock()

            # Test with commit
            dlml.download_data_files(commit='1234567', use_cache=False)
            assert mock_download.call_count == 3  # model_files.txt + 2 files
            mock_download.reset_mock()

            # Test with latest commit
            dlml.download_data_files(commit='latest', use_cache=False)
            assert mock_download.call_count == 3  # model_files.txt + 2 files


# --- Edge Case Tests ---


def test_with_empty_model_files() -> None:
    """Test with empty model_files.txt."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses
        mock_release_response = MagicMock()
        mock_release_response.json.return_value = {'target_commitish': '1234567'}

        # Mock API responses and the data directory path
        with patch('requests.get', return_value=mock_release_response), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch('pathlib.Path.open', mock_open(read_data='')), patch(
            'pelicun.tools.dlml.tqdm'
        ):
            # Download data files
            dlml.download_data_files(version='v1.0.0', use_cache=False)

            # Verify _download_file was called only for model_files.txt
            assert mock_download.call_count == 1


def test_with_large_number_of_files() -> None:
    """Test with very large number of files."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses
        mock_release_response = MagicMock()
        mock_release_response.json.return_value = {'target_commitish': '1234567'}

        # Create a large list of files
        large_file_list = '\n'.join([f'file{i}.txt' for i in range(1000)])

        # Mock API responses and the data directory path
        with patch('requests.get', return_value=mock_release_response), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch('pathlib.Path.open', mock_open(read_data=large_file_list)), patch(
            'pelicun.tools.dlml.tqdm'
        ):
            # Download data files
            dlml.download_data_files(version='v1.0.0', use_cache=False)

            # Verify _download_file was called for model_files.txt and each file in it
            assert mock_download.call_count == 1001  # model_files.txt + 1000 files


def test_with_special_characters_in_paths() -> None:
    """Test with special characters in file paths."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses
        mock_release_response = MagicMock()
        mock_release_response.json.return_value = {'target_commitish': '1234567'}

        # Create a list of files with special characters
        special_file_list = 'path/with spaces/file.txt\npath/with#hash/file.txt\npath/with?question/file.txt'

        # Mock API responses and the data directory path
        with patch('requests.get', return_value=mock_release_response), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch('pathlib.Path.open', mock_open(read_data=special_file_list)), patch(
            'pelicun.tools.dlml.tqdm'
        ):
            # Download data files
            dlml.download_data_files(version='v1.0.0', use_cache=False)

            # Verify _download_file was called for model_files.txt and each file in it
            assert mock_download.call_count == 4  # model_files.txt + 3 files

            # Verify that the URLs contain the expected file paths
            call_args_list = [call[0][0] for call in mock_download.call_args_list]
            assert any('path/with spaces/file.txt' in url for url in call_args_list)
            assert any('path/with#hash/file.txt' in url for url in call_args_list)
            assert any(
                'path/with?question/file.txt' in url for url in call_args_list
            )


def test_with_commented_lines_in_model_files() -> None:
    """Test with commented lines in model_files.txt."""
    # Create a temporary directory that will stand in for DLML_DATA_DIR
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Mock responses
        mock_release_response = MagicMock()
        mock_release_response.json.return_value = {'target_commitish': '1234567'}

        # Create a list of files with commented lines
        file_list = """# This is a comment
file1.txt
# This is another comment
file2.txt
  # This is a comment with leading whitespace
file3.txt"""

        # Mock API responses and the data directory path
        with patch('requests.get', return_value=mock_release_response), patch(
            'pelicun.tools.dlml._download_file'
        ) as mock_download, patch(
            'pelicun.tools.dlml.DLML_DATA_DIR', new=temp_dir_path
        ), patch('pathlib.Path.open', mock_open(read_data=file_list)), patch(
            'pelicun.tools.dlml.tqdm'
        ):
            # Download data files
            dlml.download_data_files(version='v1.0.0', use_cache=False)

            # Verify _download_file was called for model_files.txt and each non-commented file
            assert mock_download.call_count == 5  # model_files.txt + 4 files


# --- Integration Tests ---


def test_cache_version_tracking_integration() -> None:
    """Test integration between cache version tracking and version checking."""
    # Mock cache with version metadata
    mock_cache = {
        'version': 'v1.0.0',
        'download_type': 'version',
        'commit_sha': '1234567',
        'last_updated': '2025-08-13T10:00:00',
    }

    # Mock GitHub API response with newer version
    mock_response = MagicMock()
    mock_response.json.return_value = {'tag_name': 'v1.1.0'}
    mock_response.raise_for_status.return_value = None

    with patch('pelicun.tools.dlml.load_cache', return_value=mock_cache), patch(
        'pelicun.tools.dlml.save_cache'
    ) as mock_save, patch('requests.get', return_value=mock_response), patch(
        'pelicun.tools.dlml.DLML_DATA_DIR', new=Path('/some/mock/dir')
    ), patch('pathlib.Path.exists', return_value=True):
        result = dlml.check_dlml_version()

        # Should use cached version info for comparison
        assert result['current_version'] == 'v1.0.0'
        assert result['latest_version'] == 'v1.1.0'
        assert result['update_available'] is True

        # Should update cache with version check results
        mock_save.assert_called_once()
        updated_cache = mock_save.call_args[0][1]
        assert updated_cache['current_version'] == 'v1.0.0'
        assert updated_cache['latest_version'] == 'v1.1.0'
        assert updated_cache['update_available'] is True


# These tests are not affected by the DLML_DATA_DIR refactoring.
def test_import_integration_success() -> None:
    """Test successful import integration with check_dlml_data."""
    # Mock successful data check
    with patch('pelicun.tools.dlml.check_dlml_data') as mock_check:
        mock_check.return_value = None

        # Import should succeed without raising exceptions
        try:
            # Simulate the import process
            from pelicun.tools.dlml import check_dlml_data  # noqa: PLC0415

            check_dlml_data()
        except (ImportError, RuntimeError, ValueError) as e:
            pytest.fail(f'Import integration should not raise exception: {e}')


# These tests are not affected by the DLML_DATA_DIR refactoring.
def test_import_integration_failure() -> None:
    """Test import integration handles check_dlml_data failures."""
    # Mock failed data check
    with patch(
        'pelicun.tools.dlml.check_dlml_data',
        side_effect=RuntimeError('Download failed'),
    ):
        # Should raise ImportError as would happen in __init__.py
        from pelicun.tools.dlml import check_dlml_data  # noqa: PLC0415

        with pytest.raises(RuntimeError, match='Download failed'):
            check_dlml_data()


def test_logging_configuration() -> None:
    """Test that the module-level logger is configured and used."""
    # Verify logger exists and is configured correctly.
    assert hasattr(dlml, 'logger')
    assert dlml.logger.name == 'pelicun.dlml'

    # Test that a function uses the logger by triggering a loggable event.
    # Here, we trigger a warning in _get_changed_files by mocking a network error.
    with patch(
        'requests.get', side_effect=requests.exceptions.RequestException
    ), patch.object(dlml.logger, 'warning') as mock_warning:
        dlml._get_changed_files('base', 'head', headers={})

        # Assert that the logger's warning method was called as expected.
        mock_warning.assert_called()


def test_warning_system_integration() -> None:
    """Test integration with pelicun warning system."""

    from pelicun.pelicun_warnings import PelicunWarning  # noqa: PLC0415

    # Mock version check result with update available
    mock_version_info = {
        'update_available': True,
        'current_version': 'v1.0.0',
        'latest_version': 'v1.1.0',
        'error': None,
    }

    with patch('pathlib.Path.exists', return_value=True), patch(
        'pathlib.Path.is_dir', return_value=True
    ), patch('pathlib.Path.iterdir', return_value=[MagicMock()]), patch(
        'pelicun.tools.dlml.check_dlml_version', return_value=mock_version_info
    ), patch('warnings.warn') as mock_warn, patch(
        'pelicun.tools.dlml.DLML_DATA_DIR', new=Path('/some/mock/dir')
    ):
        dlml.check_dlml_data()

        # Should use PelicunWarning class
        mock_warn.assert_called_once()
        args, kwargs = mock_warn.call_args
        assert len(args) >= 2
        assert args[1] == PelicunWarning

        # Should have appropriate stacklevel
        assert 'stacklevel' in kwargs
        assert kwargs['stacklevel'] == 2


# --- Command-Line Interface Integration Tests ---


def test_cli_integration_invalid_action() -> None:
    """Integration test: CLI with invalid action (not 'update')."""
    # Run the CLI with invalid action
    result = subprocess.run(  # noqa: S603
        ['pelicun', 'dlml', 'invalid'],  # noqa: S607
        capture_output=True,
        text=True,
        check=False,
    )

    # Verify exit code and output
    assert result.returncode == 2  # argparse returns 2 for invalid choice
    assert "invalid choice: 'invalid'" in result.stderr


def test_cli_integration_missing_arguments() -> None:
    """Integration test: CLI with insufficient arguments."""
    # Run the CLI with no arguments for dlml
    result = subprocess.run(  # noqa: S603
        ['pelicun', 'dlml'],  # noqa: S607
        capture_output=True,
        text=True,
        check=False,
    )

    # Verify exit code and output
    assert result.returncode == 2  # argparse returns 2 for missing required argument
    assert 'required' in result.stderr



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the base module of pelicun."""

from __future__ import annotations

import argparse
import io
import platform
import re
import subprocess  # noqa: S404
import tempfile
from contextlib import redirect_stdout
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pelicun import base
from pelicun.base import ensure_value

# The tests maintain the order of definitions of the `base.py` file.


def test_options_init() -> None:
    temp_dir = tempfile.mkdtemp()

    # Create a sample user_config_options dictionary
    user_config_options = {
        'Verbose': False,
        'Seed': None,
        'LogShowMS': False,
        'LogFile': f'{temp_dir}/test_log_file',
        'PrintLog': False,
        'DemandOffset': {'PFA': -1, 'PFV': -1},
        'Sampling': {
            'SamplingMethod': 'MonteCarlo',
            'SampleSize': 1000,
            'PreserveRawOrder': False,
        },
        'SamplingMethod': 'MonteCarlo',
        'NonDirectionalMultipliers': {'ALL': 1.2},
        'EconomiesOfScale': {'AcrossFloors': True, 'AcrossDamageStates': True},
        'RepairCostAndTimeCorrelation': 0.7,
    }

    # Create an Options object using the user_config_options
    # dictionary
    options = base.Options(user_config_options)

    # Check that the Options object was created successfully
    assert options is not None

    # Check that the values of the Options object attributes match the
    # values in the user_config_options dictionary
    assert options.sampling_method == 'MonteCarlo'
    assert options.units_file is None
    assert options.demand_offset == {'PFA': -1, 'PFV': -1}
    assert options.nondir_multi_dict == {'ALL': 1.2}
    assert options.rho_cost_time == 0.7
    assert options.eco_scale == {'AcrossFloors': True, 'AcrossDamageStates': True}

    # Check that the Logger object attribute of the Options object is
    # initialized with the correct parameters
    assert options.log.verbose is False
    assert options.log.log_show_ms is False
    assert Path(ensure_value(options.log.log_file)).name == 'test_log_file'
    assert options.log.print_log is False

    # test seed property and setter
    options.seed = 42
    assert options.seed == 42

    # test rng
    assert isinstance(options.rng, np.random._generator.Generator)


def test_nondir_multi() -> None:
    options = base.Options({'NonDirectionalMultipliers': {'PFA': 1.5, 'PFV': 1.00}})
    assert options.nondir_multi_dict == {'PFA': 1.5, 'PFV': 1.0, 'ALL': 1.2}


def test_logger_init() -> None:
    # Test that the Logger object is initialized with the correct
    # attributes based on the input configuration

    temp_dir = tempfile.mkdtemp()

    log_config = {
        'verbose': True,
        'log_show_ms': False,
        'log_file': f'{temp_dir}/log.txt',
        'print_log': True,
    }
    log = base.Logger(**log_config)  # type: ignore
    assert log.verbose is True
    assert log.log_show_ms is False
    assert Path(ensure_value(log.log_file)).name == 'log.txt'
    assert log.print_log is True

    # test exceptions
    log_config = {
        'verbose': True,
        'log_show_ms': False,
        'log_file': '/',
        'print_log': True,
    }
    with pytest.raises((IsADirectoryError, FileExistsError, FileNotFoundError)):
        log = base.Logger(**log_config)  # type: ignore


def test_logger_msg() -> None:
    temp_dir = tempfile.mkdtemp()

    # Test that the msg method prints the correct message to the
    # console and log file
    log_config = {
        'verbose': True,
        'log_show_ms': True,
        'log_file': f'{temp_dir}/log.txt',
        'print_log': True,
    }
    log = base.Logger(**log_config)  # type: ignore
    # Check that the message is printed to the console
    with io.StringIO() as buf, redirect_stdout(buf):
        log.msg('This is a message')
        output = buf.getvalue()
    assert 'This is a message' in output
    # Check that the message is written to the log file
    with Path(f'{temp_dir}/log.txt').open(encoding='utf-8') as f:
        assert 'This is a message' in f.read()

    # Check if timestamp is printed
    with io.StringIO() as buf, redirect_stdout(buf):
        log.msg(
            ('This is a message\nSecond line'),
            prepend_timestamp=True,
        )
        output = buf.getvalue()
        pattern = r'(\d{2}:\d{2}:\d{2})'
        assert re.search(pattern, output) is not None


def test_logger_div() -> None:
    temp_dir = tempfile.mkdtemp()

    # We test the divider with and without the timestamp
    prepend_timestamp_args = (True, False)
    patterns = (
        r'[0-9][0-9]:[0-9][0-9]:[0-9][0-9]:[0-9][0-9][0-9][0-9][0-9][0-9]\s-+',
        r'\s+-+',
    )
    for case, pattern_str in zip(prepend_timestamp_args, patterns):
        pattern = re.compile(pattern_str)
        # Test that the div method adds a divider as intended
        log_config = {
            'verbose': True,
            'log_show_ms': True,
            'log_file': f'{temp_dir}/log.txt',
            'print_log': True,
        }
        log = base.Logger(**log_config)  # type: ignore

        # check console output
        with io.StringIO() as buf, redirect_stdout(buf):
            log.div(prepend_timestamp=case)
            output = buf.getvalue()
        assert pattern.match(output)
        # check log file
        with Path(f'{temp_dir}/log.txt').open(encoding='utf-8') as f:
            # simply check that it is not empty
            assert f.read()


@pytest.mark.skipif(
    platform.system() == 'Windows',
    reason='Skipping test on Windows due to path handling issues.',
)
def test_logger_exception() -> None:
    # Create a temporary directory for log files
    temp_dir = tempfile.mkdtemp()

    # Create a sample Python script that will raise an exception
    test_script = Path(temp_dir) / 'test_script.py'
    test_script_content = f"""
from pathlib import Path
from pelicun.base import Logger

log_file_A = Path("{temp_dir}") / 'log_A.txt'
log_file_B = Path("{temp_dir}") / 'log_B.txt'

log_A = Logger(
    log_file=log_file_A,
    verbose=True,
    log_show_ms=True,
    print_log=True,
)
log_B = Logger(
    log_file=log_file_B,
    verbose=True,
    log_show_ms=True,
    print_log=True,
)

raise ValueError('Test exception in subprocess')
"""

    # Write the test script to the file
    test_script.write_text(test_script_content)

    # Use subprocess to run the script
    process = subprocess.run(  # noqa: S603
        ['python', str(test_script)],  # noqa: S607
        capture_output=True,
        text=True,
        check=False,
    )

    # Check that the process exited with an error
    assert process.returncode == 1

    # Check the stdout/stderr for the expected output
    assert 'Test exception in subprocess' in process.stdout

    # Check that the exception was logged in the log file
    log_files = (
        Path(temp_dir) / 'log_A.txt',
        Path(temp_dir) / 'log_B.txt',
    )
    for log_file in log_files:
        assert log_file.exists(), 'Log file was not created'
        log_content = log_file.read_text()
        assert 'Test exception in subprocess' in log_content
        assert 'Traceback' in log_content
        assert 'ValueError' in log_content


def test_split_file_name() -> None:
    file_path = 'example.file.name.txt'
    name, extension = base.split_file_name(file_path)
    assert name == 'example.file.name'
    assert extension == '.txt'

    file_path = 'example'
    name, extension = base.split_file_name(file_path)
    assert name == 'example'
    assert extension == ''  # noqa: PLC1901


def test_print_system_info() -> None:
    temp_dir = tempfile.mkdtemp()

    # create a logger object
    log_config = {
        'verbose': True,
        'log_show_ms': True,
        'log_file': f'{temp_dir}/log.txt',
        'print_log': True,
    }
    log = base.Logger(**log_config)  # type: ignore

    # run print_system_info and get the console output
    with io.StringIO() as buf, redirect_stdout(buf):
        log.print_system_info()
        output = buf.getvalue()

    # verify the contents of the output
    assert 'System Information:\n' in output


def test_update_vals() -> None:
    primary = {'b': {'c': 4, 'd': 5}, 'g': 7}
    update = {'a': 1, 'b': {'c': 3, 'd': 5}, 'f': 6}
    base.update_vals(update, primary, 'update', 'primary')
    assert primary == {'b': {'c': 4, 'd': 5}, 'g': 7}  # unchanged
    assert update == {'a': 1, 'b': {'c': 3, 'd': 5}, 'f': 6, 'g': 7}  # updated
    # note: key 'g' created, 'f' left there, 'c', 'd' updated, as intended

    primary = {'a': {'b': {'c': 4}}}
    update = {'a': {'b': {'c': 3}}}
    base.update_vals(update, primary, 'update', 'primary')
    assert primary == {'a': {'b': {'c': 4}}}  # unchanged
    assert update == {'a': {'b': {'c': 3}}}  # updated

    primary = {'a': {'b': 4}}
    update = {'a': {'b': {'c': 3}}}
    with pytest.raises(ValueError, match='should not map to a dictionary'):
        base.update_vals(update, primary, 'update', 'primary')

    primary = {'a': {'b': 3}}
    update = {'a': 1, 'b': 2}
    with pytest.raises(ValueError, match='should map to a dictionary'):
        base.update_vals(update, primary, 'update', 'primary')


def test_merge_default_config() -> None:
    # Test merging an empty user config with the default config
    user_config: dict[str, object] | None = {}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == base.load_default_options()

    user_config = None  # same as {}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == base.load_default_options()

    # Test merging a user config with a single option set
    user_config = {'Verbose': True}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == {**base.load_default_options(), **user_config}

    # Test merging a user config with multiple options set
    user_config = {'Verbose': True, 'Seed': 12345}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == {**base.load_default_options(), **user_config}

    # Test merging a user config with a nested option set
    user_config = {'NonDirectionalMultipliers': {'PFA': 1.5}}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == {**base.load_default_options(), **user_config}

    # Test merging a user config with a nested option set and a top-level option set
    user_config = {'Verbose': True, 'NonDirectionalMultipliers': {'PFA': 1.5}}
    merged_config = base.merge_default_config(user_config)
    assert merged_config == {**base.load_default_options(), **user_config}


def test_convert_dtypes() -> None:
    # All columns able to be converted

    # Input DataFrame
    df_input = pd.DataFrame({'a': ['1', '2', '3'], 'b': ['4.0', '5.5', '6.75']})

    # Expected DataFrame
    df_expected = pd.DataFrame({'a': [1, 2, 3], 'b': [4.0, 5.5, 6.75]}).astype(
        {'a': 'int64', 'b': 'float64'}
    )

    # Convert data types
    df_result = base.convert_dtypes(df_input)

    pd.testing.assert_frame_equal(
        df_result, df_expected, check_index_type=False, check_column_type=False
    )

    # No columns that can be converted

    df_input = pd.DataFrame(
        {'a': ['foo', 'bar', 'baz'], 'b': ['2021-01-01', '2021-01-02', '2021-01-03']}
    )
    df_expected = df_input.copy()
    df_result = base.convert_dtypes(df_input)
    pd.testing.assert_frame_equal(
        df_result, df_expected, check_index_type=False, check_column_type=False
    )

    # Columns with mixed types

    df_input = pd.DataFrame(
        {
            'a': ['1', '2', 'three'],
            'b': ['4.0', '5.5', 'six'],
            'c': ['7', 'eight', '9'],
        }
    )
    df_result = base.convert_dtypes(df_input)
    pd.testing.assert_frame_equal(
        df_result, df_input, check_index_type=False, check_column_type=False
    )

    # None values present

    df_input = pd.DataFrame({'a': [None, '2', '3'], 'b': ['4.0', None, '6.75']})
    df_expected = pd.DataFrame({'a': [np.nan, 2, 3], 'b': [4.0, np.nan, 6.75]})
    df_result = base.convert_dtypes(df_input)
    pd.testing.assert_frame_equal(
        df_result,
        df_expected,
        check_dtype=False,
        check_index_type=False,
        check_column_type=False,
    )

    # Empty DataFrame

    df_input = pd.DataFrame({})
    df_expected = pd.DataFrame({})
    df_result = base.convert_dtypes(df_input)
    pd.testing.assert_frame_equal(
        df_result, df_expected, check_index_type=False, check_column_type=False
    )


def test_convert_to_SimpleIndex() -> None:
    # Test conversion of a multiindex to a simple index following the
    # SimCenter dash convention
    index = pd.MultiIndex.from_tuples((('a', 'b'), ('c', 'd')))
    data = pd.DataFrame([[1, 2], [3, 4]], index=index)
    data.index.names = ['name_1', 'name_2']
    data_simple = base.convert_to_SimpleIndex(data, axis=0)
    assert data_simple.index.tolist() == ['a-b', 'c-d']
    assert data_simple.index.name == '-'.join(data.index.names)

    # Test inplace modification
    df_inplace = data.copy()
    base.convert_to_SimpleIndex(df_inplace, axis=0, inplace=True)
    assert df_inplace.index.tolist() == ['a-b', 'c-d']
    assert df_inplace.index.name == '-'.join(data.index.names)

    # Test conversion of columns
    index = pd.MultiIndex.from_tuples((('a', 'b'), ('c', 'd')))
    data = pd.DataFrame([[1, 2], [3, 4]], columns=index)
    data.columns.names = ['name_1', 'name_2']
    data_simple = base.convert_to_SimpleIndex(data, axis=1)
    assert data_simple.columns.tolist() == ['a-b', 'c-d']
    assert data_simple.columns.name == '-'.join(data.columns.names)

    # Test inplace modification
    df_inplace = data.copy()
    base.convert_to_SimpleIndex(df_inplace, axis=1, inplace=True)
    assert df_inplace.columns.tolist() == ['a-b', 'c-d']
    assert df_inplace.columns.name == '-'.join(data.columns.names)

    # Test invalid axis parameter
    with pytest.raises(ValueError, match='Invalid axis parameter: 2'):
        base.convert_to_SimpleIndex(data, axis=2)


def test_convert_to_MultiIndex() -> None:
    # Test a case where the index needs to be converted to a MultiIndex
    data = pd.DataFrame({'A': (1, 2, 3), 'B': (4, 5, 6)})
    data.index = pd.Index(['A-1', 'B-1', 'C-1'])
    data_converted = base.convert_to_MultiIndex(data, axis=0, inplace=False)
    expected_index = pd.MultiIndex.from_arrays((('A', 'B', 'C'), ('1', '1', '1')))
    assert data_converted.index.equals(expected_index)
    # original data should not have changed
    assert data.index.equals(pd.Index(('A-1', 'B-1', 'C-1')))

    # Test a case where the index is already a MultiIndex
    data_converted = pd.DataFrame(
        base.convert_to_MultiIndex(data_converted, axis=0, inplace=False)
    )
    assert data_converted.index.equals(expected_index)

    # Test a case where the columns need to be converted to a MultiIndex
    data = pd.DataFrame({'A-1': (1, 2, 3), 'B-1': (4, 5, 6)})
    data_converted = base.convert_to_MultiIndex(data, axis=1, inplace=False)
    expected_columns = pd.MultiIndex.from_arrays((('A', 'B'), ('1', '1')))
    assert data_converted.columns.equals(expected_columns)
    # original data should not have changed
    assert data.columns.equals(pd.Index(('A-1', 'B-1')))

    # Test a case where the columns are already a MultiIndex
    data_converted = pd.DataFrame(
        base.convert_to_MultiIndex(data_converted, axis=1, inplace=False)
    )
    assert data_converted.columns.equals(expected_columns)

    # Test an invalid axis parameter
    with pytest.raises(ValueError, match='Invalid axis parameter: 2'):
        base.convert_to_MultiIndex(data_converted, axis=2, inplace=False)

    # inplace=True
    data = pd.DataFrame({'A': (1, 2, 3), 'B': (4, 5, 6)})
    data.index = pd.Index(['A-1', 'B-1', 'C-1'])
    base.convert_to_MultiIndex(data, axis=0, inplace=True)
    expected_index = pd.MultiIndex.from_arrays((('A', 'B', 'C'), ('1', '1', '1')))
    assert data.index.equals(expected_index)


def test_show_matrix() -> None:
    # Test with a simple 2D array
    arr = np.array(((1, 2, 3), (4, 5, 6)))
    base.show_matrix(arr)

    # Test with a DataFrame
    data = pd.DataFrame(((1, 2, 3), (4, 5, 6)), columns=('a', 'b', 'c'))
    base.show_matrix(data)

    # Test with use_describe=True
    base.show_matrix(arr, use_describe=True)


def test_multiply_factor_multiple_levels() -> None:
    # Original DataFrame definition
    data = pd.DataFrame(
        np.full((5, 3), 1.00),
        index=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
        columns=['col1', 'col2', 'col3'],
    )

    # Test 1: Basic multiplication on rows
    result_df = pd.DataFrame(
        np.array(
            [
                [2.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
            ]
        ),
        index=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
        columns=['col1', 'col2', 'col3'],
    )
    test_df = data.copy()
    base.multiply_factor_multiple_levels(test_df, {'lv1': 'A', 'lv2': 'X'}, 2)
    pd.testing.assert_frame_equal(
        test_df,
        result_df,
    )

    # Test 2: Multiplication on all rows
    result_df_all = pd.DataFrame(
        np.full((5, 3), 3.00),
        index=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
        columns=['col1', 'col2', 'col3'],
    )
    test_df = data.copy()
    base.multiply_factor_multiple_levels(test_df, {}, 3)
    pd.testing.assert_frame_equal(test_df, result_df_all)

    # Original DataFrame definition for columns test
    df_columns = pd.DataFrame(
        np.ones((3, 5)),
        index=['row1', 'row2', 'row3'],
        columns=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
    )

    # Test 3: Multiplication on columns
    result_df_columns = pd.DataFrame(
        np.array(
            [
                [2.0, 2.0, 1.0, 2.0, 1.0],
                [2.0, 2.0, 1.0, 2.0, 1.0],
                [2.0, 2.0, 1.0, 2.0, 1.0],
            ]
        ),
        index=['row1', 'row2', 'row3'],
        columns=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
    )
    test_df = df_columns.copy()
    base.multiply_factor_multiple_levels(test_df, {'lv2': 'X'}, 2, axis=1)
    pd.testing.assert_frame_equal(
        test_df,
        result_df_columns,
    )

    # Test 4: Multiplication with no matching conditions
    with pytest.raises(
        ValueError, match="No rows found matching the conditions: `{'lv1': 'C'}`"
    ):
        base.multiply_factor_multiple_levels(data.copy(), {'lv1': 'C'}, 2)

    # Test 5: Invalid axis
    with pytest.raises(ValueError, match='Invalid axis: `2`'):
        base.multiply_factor_multiple_levels(data.copy(), {'lv1': 'A'}, 2, axis=2)

    # Test 6: Empty conditions affecting all rows
    result_df_empty = pd.DataFrame(
        np.full((5, 3), 4.00),
        index=pd.MultiIndex.from_tuples(
            [
                ('A', 'X', 'K'),
                ('A', 'X', 'L'),
                ('A', 'Y', 'M'),
                ('B', 'X', 'K'),
                ('B', 'Y', 'M'),
            ],
            names=['lv1', 'lv2', 'lv3'],
        ),
        columns=['col1', 'col2', 'col3'],
    )
    testing_df = data.copy()
    base.multiply_factor_multiple_levels(testing_df, {}, 4)
    pd.testing.assert_frame_equal(testing_df, result_df_empty)


def test_describe() -> None:
    expected_idx: pd.Index = pd.Index(
        (
            'count',
            'mean',
            'std',
            'log_std',
            'min',
            '0.1%',
            '2.3%',
            '10%',
            '15.9%',
            '50%',
            '84.1%',
            '90%',
            '97.7%',
            '99.9%',
            'max',
        ),
        dtype='object',
    )

    # case 1:
    # passing a DataFrame

    data = pd.DataFrame(
        ((1.00, 2.00, 3.00), (4.00, 5.00, 6.00)), columns=['A', 'B', 'C']
    )
    desc = base.describe(data)
    assert np.all(desc.index == expected_idx)
    assert np.all(desc.columns == pd.Index(('A', 'B', 'C'), dtype='object'))

    # case 2:
    # passing a series

    sr = pd.Series((1.00, 2.00, 3.00), name='A')
    desc = base.describe(sr)
    assert np.all(desc.index == expected_idx)
    assert np.all(desc.columns == pd.Index(('A',), dtype='object'))

    # case 3:
    # passing a 2D numpy array

    desc = base.describe(np.array(((1.00, 2.00, 3.00), (4.00, 5.00, 6.00))))
    assert np.all(desc.index == expected_idx)
    assert np.all(desc.columns == pd.Index((0, 1, 2), dtype='object'))

    # case 4:
    # passing a 1D numpy array

    desc = base.describe(np.array((1.00, 2.00, 3.00)))
    assert np.all(desc.index == expected_idx)
    assert np.all(desc.columns == pd.Index((0,), dtype='object'))


def test_str2bool() -> None:
    assert base.str2bool('True') is True
    assert base.str2bool('False') is False
    assert base.str2bool('yes') is True
    assert base.str2bool('no') is False
    assert base.str2bool('t') is True
    assert base.str2bool('f') is False
    assert base.str2bool('1') is True
    assert base.str2bool('0') is False
    assert base.str2bool(v=True) is True
    assert base.str2bool(v=False) is False
    with pytest.raises(argparse.ArgumentTypeError):
        base.str2bool('In most cases, it depends..')


def test_float_or_None() -> None:
    # Test with a string that can be converted to a float
    assert base.float_or_None('123.00') == 123.00

    # Test with a string that represents an integer
    assert base.float_or_None('42') == 42.0

    # Test with a string that represents a negative number
    assert base.float_or_None('-123.00') == -123.00

    # Test with a string that can't be converted to a float
    assert base.float_or_None('hello') is None

    # Test with an empty string
    assert base.float_or_None('') is None


def test_int_or_None() -> None:
    # Test the case when the string can be converted to int
    assert base.int_or_None('123') == 123
    assert base.int_or_None('-456') == -456
    assert base.int_or_None('0') == 0
    assert base.int_or_None('+789') == 789

    # Test the case when the string cannot be converted to int
    assert base.int_or_None('abc') is None
    assert base.int_or_None('123a') is None
    assert base.int_or_None(' ') is None
    assert base.int_or_None('') is None


def test_check_if_str_is_na() -> None:
    data = ['N/A', 'foo', 'NaN', '', 'bar', np.nan]

    res = [base.check_if_str_is_na(x) for x in data]

    assert res == [True, False, True, True, False, False]


def test_with_parsed_str_na_values() -> None:
    data = pd.DataFrame(
        {
            'A': [1.00, 2.00, 'N/A', 4.00, 5.00],
            'B': ['foo', 'bar', 'NA', 'baz', 'qux'],
            'C': [1, 2, 3, 4, 5],
        }
    )

    res = base.with_parsed_str_na_values(data)
    pd.testing.assert_frame_equal(
        res,
        pd.DataFrame(
            {
                'A': [1.00, 2.00, np.nan, 4.00, 5.00],
                'B': ['foo', 'bar', np.nan, 'baz', 'qux'],
                'C': [1, 2, 3, 4, 5],
            }
        ),
    )


def test_run_input_specs() -> None:
    assert Path(base.pelicun_path).name == 'pelicun'


def test_dedupe_index() -> None:
    tuples = [('A', '1'), ('A', '1'), ('B', '2'), ('B', '3')]
    index = pd.MultiIndex.from_tuples(tuples, names=['L1', 'L2'])
    data = np.full((4, 1), 0.00)
    data_pd = pd.DataFrame(data, index=index)
    data_pd = base.dedupe_index(data_pd)
    assert data_pd.to_dict() == {
        0: {
            ('A', '1', '0'): 0.0,
            ('A', '1', '1'): 0.0,
            ('B', '2', '0'): 0.0,
            ('B', '3', '0'): 0.0,
        }
    }


def test_dict_raise_on_duplicates() -> None:
    res = base.dict_raise_on_duplicates([('A', '1'), ('B', '2')])
    assert res == {'A': '1', 'B': '2'}
    with pytest.raises(ValueError, match='duplicate key: A'):
        base.dict_raise_on_duplicates([('A', '1'), ('A', '2')])


def test_parse_units() -> None:
    # Test the default units are parsed correctly
    units = base.parse_units()
    assert isinstance(units, dict)
    expect = {
        'sec': 1.0,
        'minute': 60.0,
        'hour': 3600.0,
        'day': 86400.0,
        'm': 1.0,
        'mm': 0.001,
        'cm': 0.01,
        'km': 1000.0,
        'in': 0.0254,
        'inch': 0.0254,
        'ft': 0.3048,
        'mile': 1609.344,
        'm2': 1.0,
        'mm2': 1e-06,
        'cm2': 0.0001,
        'km2': 1000000.0,
        'in2': 0.00064516,
        'inch2': 0.00064516,
        'ft2': 0.09290304,
        'mile2': 2589988.110336,
        'm3': 1.0,
        'in3': 1.6387064e-05,
        'inch3': 1.6387064e-05,
        'ft3': 0.028316846592,
        'cmps': 0.01,
        'mps': 1.0,
        'mph': 0.44704,
        'inps': 0.0254,
        'inchps': 0.0254,
        'ftps': 0.3048,
        'mps2': 1.0,
        'inps2': 0.0254,
        'inchps2': 0.0254,
        'ftps2': 0.3048,
        'g': 9.80665,
        'kg': 1.0,
        'ton': 1000.0,
        'lb': 0.453592,
        'N': 1.0,
        'kN': 1000.0,
        'lbf': 4.4482179868,
        'kip': 4448.2179868,
        'kips': 4448.2179868,
        'Npm': 1.0,
        'kNpm': 1000.0,
        'lbpft': 14.5939,
        'kippft': 14593.9,
        'kipspft': 14593.9,
        'Pa': 1.0,
        'kPa': 1000.0,
        'MPa': 1000000.0,
        'GPa': 1000000000.0,
        'psi': 6894.751669043338,
        'ksi': 6894751.669043338,
        'Mpsi': 6894751669.043338,
        'A': 1.0,
        'V': 1.0,
        'kV': 1000.0,
        'ea': 1.0,
        'unitless': 1.0,
        'rad': 1.0,
        'C': 1.0,
        'USD_2011': 1.0,
        'USD': 1.0,
        'loss_ratio': 1.0,
        'worker_day': 1.0,
        'EA': 1.0,
        'SF': 0.09290304,
        'LF': 0.3048,
        'TN': 1000.0,
        'AP': 1.0,
        'CF': 0.0004719474432,
        'KV': 1000.0,
        'J': 1.0,
        'MJ': 1000000.0,
        'test_two': 2.00,
        'test_three': 3.00,
    }
    for thing, value in units.items():
        assert thing in expect
        assert value == expect[thing]

    # Test that additional units are parsed correctly
    additional_units_file = (
        'pelicun/tests/basic/data/base/test_parse_units/additional_units_a.json'
    )
    units = base.parse_units(additional_units_file)
    assert isinstance(units, dict)
    assert 'year' in units
    assert units['year'] == 1.00

    # Test that an exception is raised if the additional units file is not found
    with pytest.raises(FileNotFoundError):
        units = base.parse_units('invalid/file/path.json')

    # Test that an exception is raised if the additional units file is
    # not a valid JSON file
    invalid_json_file = 'pelicun/tests/basic/data/base/test_parse_units/invalid.json'
    with pytest.raises(
        ValueError,
        match='not a valid JSON file.',
    ):
        units = base.parse_units(invalid_json_file)

    # Test that an exception is raised if a unit is defined twice in
    # the additional units file
    duplicate_units_file = (
        'pelicun/tests/basic/data/base/test_parse_units/duplicate2.json'
    )
    with pytest.raises(
        ValueError,
        match='sec defined twice',
    ):
        units = base.parse_units(duplicate_units_file)

    # Test that an exception is raised if a unit conversion factor is not a float
    invalid_units_file = (
        'pelicun/tests/basic/data/base/test_parse_units/not_float.json'
    )
    with pytest.raises(TypeError):
        units = base.parse_units(invalid_units_file)

    # Test that we get an error if some first-level key does not point
    # to a dictionary
    invalid_units_file = (
        'pelicun/tests/basic/data/base/test_parse_units/not_dict.json'
    )
    with pytest.raises(
        (ValueError, TypeError),
        match="contains first-level keys that don't point to a dictionary",
    ):
        units = base.parse_units(invalid_units_file)


def test_unit_conversion() -> None:
    # Test scalar conversion from feet to meters
    assert base.convert_units(1.00, 'ft', 'm') == 0.3048

    # Test list conversion from feet to meters
    feet_values_list = [1.0, 2.0, 3.0]
    meter_values_list = [0.3048, 0.6096, 0.9144]
    np.testing.assert_array_almost_equal(
        base.convert_units(feet_values_list, 'ft', 'm'), meter_values_list
    )

    # Test numpy array conversion from feet to meters
    feet_values_array = np.array([1.0, 2.0, 3.0])
    meter_values_array = np.array([0.3048, 0.6096, 0.9144])
    np.testing.assert_array_almost_equal(
        base.convert_units(feet_values_array, 'ft', 'm'), meter_values_array
    )

    # Test conversion with explicit category
    assert base.convert_units(1.00, 'ft', 'm', category='length') == 0.3048

    # Test error handling for invalid input type
    with pytest.raises(TypeError) as excinfo:
        base.convert_units('one', 'ft', 'm')  # type: ignore
    assert str(excinfo.value) == 'Invalid input type for `values`'

    # Test error handling for unknown unit
    with pytest.raises(ValueError, match='Unknown unit `xyz`'):
        base.convert_units(1.00, 'xyz', 'm')

    # Test error handling for mismatched category
    with pytest.raises(ValueError, match='Unknown unit: `ft`'):
        base.convert_units(1.00, 'ft', 'm', category='volume')

    # Test error handling unknown category
    with pytest.raises(ValueError, match='Unknown category: `unknown_category`'):
        base.convert_units(1.00, 'ft', 'm', category='unknown_category')

    # Test error handling different categories
    with pytest.raises(
        ValueError,
        match='`lb` is a `mass` unit, but `m` is not specified in that category.',
    ):
        base.convert_units(1.00, 'lb', 'm')


def test_stringterpolation() -> None:
    func = base.stringterpolation('1,2,3|4,5,6')
    x_new = np.array([4, 4.5, 5])
    expected = np.array([1, 1.5, 2])
    np.testing.assert_array_almost_equal(func(x_new), expected)


def test_invert_mapping() -> None:
    original_dict = {'a': [1, 2], 'b': [3]}
    expected = {1: 'a', 2: 'a', 3: 'b'}
    assert base.invert_mapping(original_dict) == expected

    # with duplicates, raises an error
    original_dict = {'a': [1, 2], 'b': [2]}
    with pytest.raises(
        ValueError, match='Cannot invert mapping with duplicate values.'
    ):
        base.invert_mapping(original_dict)



# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

"""
Tests a complete loss estimation workflow combining damage state and
loss function driven components.
The code is based on PRJ-3411v5 hosted on DesignSafe.

"""

import tempfile

import numpy as np
import pandas as pd
import pytest

import pelicun
from pelicun import assessment, file_io
from pelicun.pelicun_warnings import PelicunWarning


def test_combined_workflow() -> None:
    temp_dir = tempfile.mkdtemp()

    sample_size = 10000

    # Initialize a pelicun assessment
    asmnt = assessment.Assessment(
        {'PrintLog': True, 'Seed': 415, 'LogFile': f'{temp_dir}/log_file.txt'}
    )

    asmnt.options.list_all_ds = True
    asmnt.options.eco_scale['AcrossFloors'] = True
    asmnt.options.eco_scale['AcrossDamageStates'] = True

    demand_data = file_io.load_data(
        'pelicun/tests/validation/v2/data/demand_data.csv',
        unit_conversion_factors=None,
        reindex=False,
    )
    ndims = len(demand_data)
    perfect_correlation = pd.DataFrame(
        np.ones((ndims, ndims)),
        columns=demand_data.index,  # type: ignore
        index=demand_data.index,  # type: ignore
    )

    #
    # Additional damage state-driven components
    #

    damage_db = pelicun.file_io.load_data(
        'pelicun/tests/validation/v2/data/additional_damage_db.csv',
        reindex=False,
        unit_conversion_factors=asmnt.unit_conversion_factors,
    )
    consequences = pelicun.file_io.load_data(
        'pelicun/tests/validation/v2/data/additional_consequences.csv',
        reindex=False,
        unit_conversion_factors=asmnt.unit_conversion_factors,
    )

    #
    # Additional loss function-driven components
    #

    loss_functions = pelicun.file_io.load_data(
        'pelicun/tests/validation/v2/data/additional_loss_functions.csv',
        reindex=False,
        unit_conversion_factors=asmnt.unit_conversion_factors,
    )

    #
    # Demands
    #

    # Load the demand model
    asmnt.demand.load_model(
        {'marginals': demand_data, 'correlation': perfect_correlation}
    )

    # Generate samples
    asmnt.demand.generate_sample({'SampleSize': sample_size})

    def add_more_edps() -> None:
        """Adds SA_1.13 and residual drift to the demand sample."""
        # Add residual drift and Sa
        demand_sample = asmnt.demand.save_sample()

        # RIDs are all fixed for testing.
        rid = pd.concat(
            [
                pd.DataFrame(
                    np.full(demand_sample['PID'].shape, 0.0050),  # type: ignore
                    index=demand_sample['PID'].index,  # type: ignore
                    columns=demand_sample['PID'].columns,  # type: ignore
                )
            ],
            axis=1,
            keys=['RID'],
        )
        demand_sample_ext = pd.concat([demand_sample, rid], axis=1)  # type: ignore

        demand_sample_ext['SA_1.13', 0, 1] = 1.50

        # Add units to the data
        demand_sample_ext.T.insert(0, 'Units', '')

        # PFA and SA are in "g" in this example, while PID and RID are "rad"
        demand_sample_ext.loc['Units', ['PFA', 'SA_1.13']] = 'g'
        demand_sample_ext.loc['Units', ['PID', 'RID']] = 'rad'

        asmnt.demand.load_sample(demand_sample_ext)

    add_more_edps()

    #
    # Asset
    #

    # Specify number of stories
    asmnt.stories = 1

    # Load component definitions
    cmp_marginals = pd.read_csv(
        'pelicun/tests/validation/v2/data/CMP_marginals.csv', index_col=0
    )
    cmp_marginals['Blocks'] = cmp_marginals['Blocks']
    asmnt.asset.load_cmp_model({'marginals': cmp_marginals})

    # Generate sample
    asmnt.asset.generate_cmp_sample(sample_size)

    # # Used to test that the example works when an existing sample is
    # # loaded.
    # asmnt.asset.save_cmp_sample(filepath='/tmp/cmp_sample.csv', save_units=True)
    # asmnt.asset.cmp_sample
    # asmnt.asset.load_cmp_sample(filepath='/tmp/cmp_sample.csv')
    # asmnt.asset.cmp_sample

    #
    # Damage
    #

    cmp_set = set(asmnt.asset.list_unique_component_ids())

    # Load the models into pelicun
    asmnt.damage.load_model_parameters(
        [
            damage_db,  # type: ignore
            'PelicunDefault/FEMA P-58/fragility.csv',
        ],
        cmp_set,
    )

    # Prescribe the damage process
    dmg_process = {
        '1_collapse': {'DS1': 'ALL_NA'},
        '2_excessiveRID': {'DS1': 'irreparable_DS1'},
    }

    # Calculate damages

    asmnt.damage.calculate(dmg_process=dmg_process)

    # Test load sample, save sample
    asmnt.damage.save_sample(f'{temp_dir}/out.csv')
    asmnt.damage.load_sample(f'{temp_dir}/out.csv')

    assert asmnt.damage.ds_model.sample is not None
    asmnt.damage.ds_model.sample.mean()

    #
    # Losses
    #

    # Create the loss map
    loss_map = pd.DataFrame(
        ['replacement', 'replacement'],
        columns=['Repair'],
        index=['collapse', 'irreparable'],
    )

    # Load the loss model
    asmnt.loss.decision_variables = ('Cost', 'Time')
    asmnt.loss.add_loss_map(loss_map, loss_map_policy='fill')
    with pytest.warns(PelicunWarning):
        asmnt.loss.load_model_parameters(
            [
                consequences,  # type: ignore
                loss_functions,  # type: ignore
                'PelicunDefault/FEMA P-58/consequence_repair.csv',
            ]
        )

    # Perform the calculation
    asmnt.loss.calculate()

    # Test load sample, save sample
    with pytest.warns(PelicunWarning):
        asmnt.loss.save_sample(f'{temp_dir}/sample.csv')
        asmnt.loss.load_sample(f'{temp_dir}/sample.csv')

    #
    # Loss sample aggregation
    #

    # Get the aggregated losses
    with pytest.warns(PelicunWarning):
        agg_df = asmnt.loss.aggregate_losses()
    assert agg_df is not None
